"""X/Y fitting on layers."""

from __future__ import annotations

import weakref
from typing import TYPE_CHECKING, Generic, TypeVar

import numpy as np
from numpy.typing import NDArray

from whitecanvas import theme
from whitecanvas._exceptions import ReferenceDeletedError
from whitecanvas.layers import InfCurve, InfLine, Layer
from whitecanvas.layers._base import DataBoundLayer
from whitecanvas.protocols.layer_protocols import XYDataProtocol
from whitecanvas.types import ColorType, LineStyle, XYData

if TYPE_CHECKING:
    from whitecanvas.canvas._base import CanvasBase

_C = TypeVar("_C", bound="CanvasBase")
_L = TypeVar("_L", bound=Layer)
_P = TypeVar("_P", bound=XYDataProtocol)


class FitPlotter(Generic[_C, _P]):
    def __init__(self, canvas: _C, layer: DataBoundLayer[_P, XYData]):
        self._canvas_ref = weakref.ref(canvas)
        self._layer_ref = weakref.ref(layer)

    def _canvas(self) -> _C:
        canvas = self._canvas_ref()
        if canvas is None:
            raise ReferenceDeletedError("Canvas has been deleted.")
        return canvas

    def _layer(self) -> DataBoundLayer[_P, XYData]:
        layer = self._layer_ref()
        if layer is None:
            raise ReferenceDeletedError("Layer has been deleted.")
        return layer

    def linear(
        self,
        *,
        color: ColorType | None = None,
        width: float | None = None,
        style: str | LineStyle | None = None,
    ) -> InfLine:
        """
        Add a linear regression result of the X/Y data.

        >>> layer = canvas.add_markers(x, y)  # add markers
        >>> canvas.fit(layer).linear()  # linear regressions

        Parameters
        ----------
        color : color-like, optional
            Color of the output line.
        width : float, optional
            Width of the output line.
        style : str or LineStyle, optional
            Line style of the output line.

        Returns
        -------
        InfLine
            Infinite line representing the linear regression result.
        """
        layer = self._layer()
        canvas = self._canvas()
        data = layer.data

        def _reg(x: NDArray[np.number], y: NDArray[np.number]):
            n = len(x)
            xs1 = x.sum()
            ys1 = y.sum()
            xs2 = (x**2).sum()
            xydot = np.dot(x, y)
            a = (xydot - ys1 * xs1 / n) / (xs2 - xs1**2 / n)
            b = (ys1 - a * xs1) / n
            return a, b

        a, b = _reg(data.x, data.y)
        color = canvas._generate_colors(color)
        width = theme._default("line.width", width)
        style = theme._default("line.style", style)
        new = InfLine(
            (0, b), angle=np.rad2deg(np.arctan(a)), color=color, width=width,
            style=style, name=f"Fit<{layer.name}>", backend=canvas._get_backend(),
        )  # fmt: skip
        return canvas.add_layer(new)

    def polynomial(
        self,
        degree: int,
        *,
        color: ColorType | None = None,
        width: float | None = None,
        style: str | LineStyle | None = None,
    ) -> InfCurve:
        """
        Add a polynomial regression result of any degree of the X/Y data.

        >>> layer = canvas.add_markers(x, y)  # add markers
        >>> canvas.fit(layer).polynomial(degree=2)  # fit to x^2 + bx + c

        Parameters
        ----------
        degree : int
            Degree of the polynomial.
        color : color-like, optional
            Color of the output line.
        width : float, optional
            Width of the output line.
        style : str or LineStyle, optional
            Line style of the output line.

        Returns
        -------
        InfLine
            Infinite line representing the linear regression result.
        """
        layer = self._layer()
        canvas = self._canvas()
        data = layer.data
        poly: np.polynomial.Polynomial = np.polynomial.Polynomial.fit(
            data.x, data.y, degree
        )
        color = canvas._generate_colors(color)
        width = theme._default("line.width", width)
        style = theme._default("line.style", style)
        new = InfCurve(
            poly, color=color, width=width, style=style, name=f"Fit<{layer.name}>",
            backend=canvas._get_backend(),
        )  # fmt: skip
        return canvas.add_layer(new)



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, cast

import numpy as np
import pyqtgraph as pg
import qtpy
from pyqtgraph.GraphicsScene.mouseEvents import HoverEvent as pgHoverEvent
from pyqtgraph.GraphicsScene.mouseEvents import MouseClickEvent as pgMouseClickEvent
from qtpy import QtCore, QtGui
from qtpy.QtCore import Signal

from whitecanvas import protocols
from whitecanvas.backend.pyqtgraph._base import InsetPlotItem, PyQtAxis, PyQtLayer
from whitecanvas.backend.pyqtgraph._labels import Axis, AxisLabel, Ticks, Title
from whitecanvas.backend.pyqtgraph._legend import QtItemSampleBase, make_sample_item
from whitecanvas.backend.pyqtgraph._qt_utils import from_qt_button, from_qt_modifiers
from whitecanvas.layers._legend import LegendItem, LegendItemCollection
from whitecanvas.types import (
    Location,
    MouseButton,
    MouseEvent,
    MouseEventType,
    Rect,
)

if TYPE_CHECKING:
    from whitecanvas.canvas import CanvasBase


@protocols.check_protocol(protocols.CanvasProtocol)
class Canvas:
    def __init__(
        self,
        item: pg.PlotItem,
        *,
        xaxis: str = "bottom",
        yaxis: str = "left",
    ):
        item.vb.disableAutoRange()  # auto range is done in the whitecanvas side
        self._signals = SignalListener()
        item.vb.sigRangeChanged.connect(self._signals._set_rect)
        item.addItem(self._signals)
        self._plot_item = item
        self._xaxis = Axis(self, axis=xaxis)
        self._yaxis = Axis(self, axis=yaxis)
        self._xticks = Ticks(self, axis=xaxis)
        self._yticks = Ticks(self, axis=yaxis)
        self._title = Title(self)
        self._xlabel = AxisLabel(self, axis=xaxis)
        self._ylabel = AxisLabel(self, axis=yaxis)

    def _plt_get_native(self):
        return self._plot_item

    def _plt_get_title(self):
        return self._title

    def _plt_get_xaxis(self):
        return self._xaxis

    def _plt_get_yaxis(self):
        return self._yaxis

    def _plt_get_xlabel(self):
        return self._xlabel

    def _plt_get_xticks(self):
        return self._xticks

    def _plt_get_yticks(self):
        return self._yticks

    def _plt_get_ylabel(self):
        return self._ylabel

    def _plt_reorder_layers(self, layers: list[PyQtLayer]):
        for i, layer in enumerate(layers):
            layer._plt_set_zorder(i)

    def _plt_get_aspect_ratio(self) -> float | None:
        """Get aspect ratio of canvas"""
        locked = self._viewbox().state["aspectLocked"]
        if not locked:
            return None
        return float(locked)

    def _plt_set_aspect_ratio(self, ratio: float | None):
        """Set aspect ratio of canvas"""
        if ratio is None:
            self._viewbox().setAspectLocked(lock=False)
        else:
            self._viewbox().setAspectLocked(lock=True, ratio=ratio)

    def _viewbox(self) -> pg.ViewBox:
        return self._plot_item.vb

    def _plt_add_layer(self, layer: protocols.BaseProtocol):
        self._plot_item.addItem(layer)

    def _plt_remove_layer(self, layer):
        """Remove layer from the canvas"""
        self._plot_item.removeItem(layer)

    def _plt_get_visible(self) -> bool:
        """Get visibility of canvas"""
        return self._plot_item.isVisible()

    def _plt_set_visible(self, visible: bool):
        """Set visibility of canvas"""
        self._plot_item.setVisible(visible)

    def _plt_twinx(self) -> Canvas:
        """Create a twinx canvas"""
        vb1 = self._plot_item.vb
        vb2 = pg.ViewBox()
        new_item = pg.PlotItem(
            viewBox=vb2,
            axisItems={"right": PyQtAxis("right"), "bottom": PyQtAxis("bottom")},
        )
        canvas = Canvas(new_item, yaxis="right")

        self._get_scene().addItem(vb2)
        self._plot_item.getAxis("right").linkToView(vb2)
        self._plot_item.showAxis("right")
        vb2.setXLink(self._plot_item)

        def _update_views():
            vb2.setGeometry(vb1.sceneBoundingRect())
            vb2.linkedViewChanged(vb1, vb2.XAxis)

        _update_views()
        vb1.sigResized.connect(_update_views)

        return canvas

    def _plt_twiny(self) -> Canvas:
        vb1 = self._plot_item.vb
        vb2 = pg.ViewBox()
        new_item = pg.PlotItem(
            viewBox=vb2, axisItems={"left": PyQtAxis("left"), "top": PyQtAxis("top")}
        )
        canvas = Canvas(new_item, xaxis="top")

        self._get_scene().addItem(vb2)
        self._plot_item.getAxis("bottom").linkToView(vb2)
        self._plot_item.showAxis("top")
        vb2.setYLink(self._plot_item)

        def _update_views():
            vb2.setGeometry(vb1.sceneBoundingRect())
            vb2.linkedViewChanged(vb1, vb2.YAxis)

        _update_views()
        vb1.sigResized.connect(_update_views)

        return canvas

    def _plt_inset(self, rect: Rect) -> Canvas:
        vb = pg.ViewBox()
        item = pg.PlotItem(
            viewBox=vb,
            axisItems={"left": PyQtAxis("left"), "bottom": PyQtAxis("bottom")},
        )
        canvas = Canvas(item)
        inset_item = InsetPlotItem(item, rect)
        inset_item.setParentItem(self._plot_item.vb)
        return canvas

    def _get_scene(self) -> pg.GraphicsScene:
        return self._plot_item.scene()

    def _plt_connect_mouse_click(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""

        def _cb(ev):
            mev = self._translate_mouse_event(ev, MouseEventType.PRESS)
            callback(mev)

        self._signals.pressed.connect(_cb)

    def _plt_connect_mouse_drag(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""

        def _cb(qpoint: QtCore.QPointF):
            mev = self._translate_mouse_event(qpoint, MouseEventType.MOVE)
            callback(mev)
            if not self._plt_get_mouse_enabled():
                # for some reason, plot items are not updated when mouse is disabled
                self._viewbox().update()

        self._signals.moved.connect(_cb)

    def _plt_connect_mouse_double_click(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""

        def _cb(ev):
            mev = self._translate_mouse_event(ev, MouseEventType.DOUBLE_CLICK)
            callback(mev)

        self._signals.double_clicked.connect(_cb)

    def _plt_connect_mouse_release(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""

        def _cb(ev):
            mev = self._translate_mouse_event(ev, MouseEventType.RELEASE)
            callback(mev)

        self._signals.released.connect(_cb)

    def _plt_connect_xlim_changed(
        self, callback: Callable[[tuple[float, float]], None]
    ):
        """Connect callback to x-limits changed event"""
        self._plot_item.sigXRangeChanged.connect(lambda _, x: callback(x))

    def _plt_connect_ylim_changed(
        self, callback: Callable[[tuple[float, float]], None]
    ):
        """Connect callback to y-limits changed event"""
        self._plot_item.sigYRangeChanged.connect(lambda _, y: callback(y))

    def _plt_draw(self):
        pass  # pyqtgraph has its own draw mechanism

    def _plt_make_legend(
        self,
        items: list[tuple[str, LegendItem]],
        anchor: Location,
    ):
        pos, offset = _LEGEND_POS[anchor]
        legend = self._plot_item.addLegend(sampleType=QtItemSampleBase)
        legend.anchor(pos, pos, offset=offset)
        for label, item in items:
            if item is None:
                continue
            if isinstance(item, LegendItemCollection):
                for sub_label, sub_item in item.items:
                    sample = make_sample_item(sub_item)
                    if sample is not None:
                        legend.addItem(sample, sub_label)
            else:
                sample = make_sample_item(item)
                if sample is not None:
                    legend.addItem(sample, label)

    def _plt_get_mouse_enabled(self):
        return self._viewbox().mouseEnabled()[0]

    def _plt_set_mouse_enabled(self, enabled: bool):
        self._viewbox().setMouseEnabled(enabled, enabled)

    def _translate_mouse_event(
        self,
        ev: QtGui.QMouseEvent | QtCore.QPointF,
        typ: MouseEventType,
    ) -> MouseEvent:
        """Translate a mouse event from pyqtgraph to whitecanvas."""
        if isinstance(ev, QtCore.QPointF):
            # Hover events only have positions.
            qpoint = ev
            scene = self._get_scene()
            btns = scene.dragButtons
            evs: list[pgMouseClickEvent] = scene.clickEvents
            if len(btns) == 0:
                button = MouseButton.NONE
            else:
                button = from_qt_button(btns[0])
            if len(evs) == 0:
                modifiers = ()
            else:
                modifiers = from_qt_modifiers(evs[0].modifiers())
            mev = MouseEvent(
                button=button,
                modifiers=modifiers,
                pos=(qpoint.x(), qpoint.y()),
                type=typ,
            )
        else:
            qpoint = cast(QtCore.QPointF, ev.pos())

            modifiers = from_qt_modifiers(ev.modifiers())

            mev = MouseEvent(
                button=from_qt_button(ev.button()),
                modifiers=modifiers,
                pos=(qpoint.x(), qpoint.y()),
                type=typ,
            )
        return mev

    def _plt_canvas_hook(self, canvas: CanvasBase):
        self._plot_item.autoBtn.clicked.disconnect()
        self._plot_item.autoBtn.clicked.connect(lambda: canvas.autoscale())


_LEGEND_POS = {
    Location.TOP_LEFT: ((0.0, 0.0), (10, 10)),
    Location.TOP_CENTER: ((0.5, 0.0), (0, 10)),
    Location.TOP_RIGHT: ((1.0, 0.0), (-10, 10)),
    Location.CENTER_LEFT: ((0.0, 0.5), (10, 0)),
    Location.CENTER: ((0.5, 0.5), (0, 0)),
    Location.CENTER_RIGHT: ((1.0, 0.5), (-10, 0)),
    Location.BOTTOM_LEFT: ((0.0, 1.0), (10, -10)),
    Location.BOTTOM_CENTER: ((0.5, 1.0), (0, -10)),
    Location.BOTTOM_RIGHT: ((1.0, 1.0), (-10, -10)),
    # These are not supported in pyqtgraph. Use the closest one.
    Location.LEFT_SIDE_TOP: ((0.0, 0.0), (10, 10)),
    Location.LEFT_SIDE_CENTER: ((0.0, 0.5), (10, 0)),
    Location.LEFT_SIDE_BOTTOM: ((0.0, 1.0), (10, -10)),
    Location.RIGHT_SIDE_TOP: ((1.0, 0.0), (-10, 10)),
    Location.RIGHT_SIDE_CENTER: ((1.0, 0.5), (-10, 0)),
    Location.RIGHT_SIDE_BOTTOM: ((1.0, 1.0), (-10, -10)),
    Location.TOP_SIDE_LEFT: ((0.0, 0.0), (10, 10)),
    Location.TOP_SIDE_CENTER: ((0.5, 0.0), (0, 10)),
    Location.TOP_SIDE_RIGHT: ((1.0, 0.0), (-10, 10)),
    Location.BOTTOM_SIDE_LEFT: ((0.0, 1.0), (10, -10)),
    Location.BOTTOM_SIDE_CENTER: ((0.5, 1.0), (0, -10)),
    Location.BOTTOM_SIDE_RIGHT: ((1.0, 1.0), (-10, -10)),
}


@protocols.check_protocol(protocols.CanvasGridProtocol)
class CanvasGrid:
    def __init__(self, heights: list[float], widths: list[float], app: str = "default"):
        if app == "notebook":
            from pyqtgraph.jupyter import GraphicsLayoutWidget
        elif app in ("default", "qt"):
            from pyqtgraph import GraphicsLayoutWidget
        else:
            raise ValueError(f"pyqtgraph does not support {app!r}")
        self._layoutwidget = GraphicsLayoutWidget()
        self._heights = heights
        self._widths = widths

    def _plt_add_canvas(self, row: int, col: int, rowspan: int, colspan: int) -> Canvas:
        vb = pg.ViewBox()
        item = pg.PlotItem(viewBox=vb)
        rspan = sum(self._heights[row : row + rowspan])
        cspan = sum(self._widths[col : col + colspan])
        r = sum(self._heights[:row])
        c = sum(self._widths[:col])
        self._layoutwidget.addItem(item, r, c)
        if rspan != 1:
            self._layoutwidget.ci.layout.setRowStretchFactor(row, rspan)
        if cspan != 1:
            self._layoutwidget.ci.layout.setColumnStretchFactor(col, cspan)
        canvas = Canvas(item)
        return canvas

    def _plt_show(self) -> bool:
        self._layoutwidget.setVisible(True)

    def _get_background_brush(self) -> QtGui.QBrush:
        return self._layoutwidget.backgroundBrush()

    def _plt_get_background_color(self):
        brush = self._get_background_brush()
        return np.array(brush.color().getRgbF())

    def _plt_set_background_color(self, color):
        brush = self._get_background_brush()
        brush.setColor(QtGui.QColor.fromRgbF(*color))
        self._layoutwidget.setBackgroundBrush(brush)

    def _plt_screenshot(self):
        img: QtGui.QImage = self._layoutwidget.grab().toImage()
        bits = img.constBits()
        h, w, c = img.height(), img.width(), 4
        if qtpy.API_NAME.startswith("PySide"):
            arr = np.asarray(bits).reshape(h, w, c)
        else:
            bits.setsize(h * w * c)
            arr = np.frombuffer(bits, np.uint8).reshape(h, w, c)

        return arr[:, :, [2, 1, 0, 3]]

    def _plt_set_figsize(self, width: int, height: int):
        self._layoutwidget.resize(width, height)

    def _plt_set_spacings(self, wspace: float, hspace: float):
        self._layoutwidget.ci.layout.setHorizontalSpacing(wspace)
        self._layoutwidget.ci.layout.setVerticalSpacing(hspace)


class SignalListener(pg.GraphicsObject):
    # Mouse events in pyqtgraph is very complicated.
    # Adding this graphics object to the scene will make it easier.
    pressed = Signal(object)
    moved = Signal(object)
    released = Signal(object)
    double_clicked = Signal(object)

    def __init__(self):
        super().__init__()
        self._rect = QtCore.QRectF(-10000, -10000, 10000, 10000)
        self._had_button = False

    def _set_rect(self, vb, rng: tuple[tuple[float, float], tuple[float, float]], _):
        (x0, x1), (y0, y1) = rng
        self._rect = QtCore.QRectF(x0, y0, x1 - x0, y1 - y0)

    def boundingRect(self):
        return self._rect

    def mousePressEvent(self, ev: pgMouseClickEvent):
        self.pressed.emit(ev)
        self._had_button = bool(ev.buttons() ^ QtCore.Qt.MouseButton.NoButton)
        super().mousePressEvent(ev)

    def hoverEvent(self, ev: pgHoverEvent):
        if ev.isExit():
            return
        btns: QtCore.Qt.MouseButtons = ev.buttons()

        has_button = bool(btns ^ QtCore.Qt.MouseButton.NoButton)
        if (not has_button) and self._had_button:
            self.released.emit(ev.pos())
        else:
            self.moved.emit(ev.pos())
        self._had_button = has_button

    def mouseDoubleClickEvent(self, ev):
        self.double_clicked.emit(ev)
        super().mouseDoubleClickEvent(ev)

    def paint(self, *args):
        pass



import pytest
from whitecanvas import new_canvas
import numpy as np


def test_hist():
    canvas = new_canvas(backend="mock")
    rng = np.random.default_rng(1642)
    data = np.concatenate([rng.normal(size=50), rng.normal(1.2, size=40)])
    canvas.add_hist(data, bins="auto")
    canvas.add_hist(data, bins=10)
    canvas.add_hist(data, bins=10, limits=(-2, 2))
    canvas.add_hist(data, bins=[-2, 0, 1, 2], shape="step")
    canvas.add_hist(data, bins=[-2, 0, 1, 2], shape="polygon")
    canvas.add_hist(data, bins=[-2, 0, 1, 2], shape="bars")
    with pytest.raises(ValueError):
        canvas.add_hist(data, bins=[-2, 0, 1, 2], shape="not-a-shape")
    canvas.add_hist(data, bins=[-2, -1, 0, 1.4, 2.5], kind="density")
    canvas.add_hist(data, bins=[-2, -1, 0, 1.4, 2.5], kind="probability")
    canvas.add_hist(data, bins=[-2, -1, 0, 1.4, 2.5], kind="frequency")
    layer = canvas.add_hist(data, bins=[-2, -1, 0, 1.4, 2.5], kind="percent")
    layer.data
    layer.data = data + 1
    layer.shape
    layer.shape = "step"
    layer.shape = "polygon"
    layer.shape = "bars"
    layer.zorders
    layer.zorders = [0, 1]
    layer.kind
    layer.kind = "density"
    layer.limits
    layer.limits = (-2, 2)
    layer.edges
    layer.update_edges([-2, 0, 1, 2])

def test_kde():
    canvas = new_canvas(backend="mock")
    rng = np.random.default_rng(1642)
    data = np.concatenate([rng.normal(size=50), rng.normal(1.2, size=40)])
    canvas.add_kde(data)
    layer = canvas.add_kde(data, band_width=0.6)
    layer.bottom
    layer.bottom = 1
    layer.band_width
    layer.band_width = 0.5

def test_hover_template():
    canvas = new_canvas(backend="mock")
    layer = canvas.add_markers([1, 2, 3], [4, 5, 6])
    layer.with_hover_template("const-text")
    layer.with_hover_template("{x}")
    layer.with_hover_template("{x}, {y}")
    layer.with_hover_template("{i}")
    layer.with_hover_template("a: {a}", extra={"a": [0, 1, 4]})
    layer.with_hover_template("x: {x}", extra={"x": [0, 1, 4]})
    layer.with_hover_template("a: {a:.2f}", extra={"a": [0, 1, 4.0], "b": [1, 2, 3]})
    with pytest.raises(KeyError):
        layer.with_hover_template("c: {c}", extra={"a": [0, 1, 4], "b": [1, 2, 3]})
    with pytest.raises(ValueError):
        layer.with_hover_template("c: {c}", extra={"c": [0, 1, 4, 5]})

def test_fit():
    canvas = new_canvas(backend="mock")
    rng = np.random.default_rng(1642)
    x = rng.normal(size=50)
    y = rng.normal(size=50) + 2 * x
    layer = canvas.add_markers(x, y)
    canvas.fit(layer).linear(color="red")
    canvas.fit(layer).polynomial(2)

def test_between():
    canvas = new_canvas(backend="mock")
    layer0 = canvas.add_markers([0] * 5, [0, 1, 2, 3, 4])
    layer1 = canvas.add_markers([1] * 5, [5, 1, 4, 2, 4])
    layer2 = canvas.add_markers([2] * 4, [0, 1, 2, 3])
    canvas.between(layer0, layer1).connect_points()
    with pytest.raises(ValueError):
        canvas.between(layer0, layer2).connect_points()

def test_melt():
    canvas = new_canvas(backend="mock")
    rng = np.random.default_rng(1642)
    df = {"x": rng.normal(size=15), "y": rng.normal(size=15)}
    canvas.cat_x(df).melt().add_boxplot(color="variable")

def test_sort():
    rng = np.random.default_rng(1642)
    df = {
        "x": ["a"] * 14 + ["b"] * 10 + ["c"] * 16,
        "y": ["p"] * 20 + ["q"] * 20,
        "val": rng.normal(size=40),
    }

    # simple
    canvas = new_canvas(backend="mock")
    canvas.cat_x(df, "x", "val").sort_in_order(["b", "a", "c"]).add_boxplot()
    assert canvas.x.ticks.labels == ["b", "a", "c"]

    # with color
    canvas = new_canvas(backend="mock")
    canvas.cat_x(df, "x", "val").sort_in_order(["b", "a", "c"]).add_boxplot(color="y")
    assert canvas.x.ticks.labels == ["b", "a", "c"]

    # ascending sort
    canvas = new_canvas(backend="mock")
    canvas.cat_x(df, "x", "val").sort(ascending=True).add_stripplot()
    assert canvas.x.ticks.labels == ["a", "b", "c"]

    # descending sort
    canvas = new_canvas(backend="mock")
    canvas.cat_x(df, "x", "val").sort(ascending=False).add_stripplot()
    assert canvas.x.ticks.labels == ["c", "b", "a"]

    # two keys
    canvas = new_canvas(backend="mock")
    (
        canvas
        .cat_y(df, "val", ["x", "y"])
        .sort_in_order([("b", "q"), ("a", "p"), ("c", "q"), ("c", "p"), ("a", "q"), ("b", "p")])
        .add_swarmplot()
    )
    assert canvas.y.ticks.labels == ["b\nq", "a\np", "c\nq", "c\np", "a\nq", "b\np"]

    # xy categorical plotter
    canvas = new_canvas(backend="mock")
    (
        canvas
        .cat_xy(df, "x", "y")
        .sort_in_order(x=["b", "a", "c"], y=["q", "p"])
        .mean()
        .add_heatmap("val")
    )
    assert canvas.x.ticks.labels == ["b", "a", "c"]
    assert canvas.y.ticks.labels == ["q", "p"]

def test_matplotlib_tooltip():
    canvas = new_canvas(backend="matplotlib")
    canvas.add_markers([1, 2, 3], [4, 5, 6])
    canvas._canvas()._set_tooltip((2, 5), "tooltip")
    canvas._canvas()._hide_tooltip()



import numpy as np
from whitecanvas import new_canvas
from whitecanvas import tools
from unittest.mock import MagicMock

def test_line_selector(backend: str):
    canvas = new_canvas(backend)
    mock = MagicMock()
    selector = tools.line_selector(canvas)
    selector.changed.connect(mock)
    canvas.mouse.emulate_drag([[0, 0], [0, 1], [1, 1]])
    assert mock.call_count == 1
    mock.assert_called_with(((0, 0), (1, 1)))

    canvas = new_canvas(backend)
    mock = MagicMock()
    selector = tools.line_selector(canvas, tracking=True)
    selector.changed.connect(mock)
    canvas.mouse.emulate_drag([[0, 0], [0, 1], [1, 1]])
    assert mock.call_count == 2

def test_rect_selector(backend: str):
    canvas = new_canvas(backend)
    mock = MagicMock()
    selector = tools.rect_selector(canvas)
    selector.changed.connect(mock)
    canvas.mouse.emulate_drag([[0, 0], [0, 1], [1, 1]])
    assert mock.call_count == 1

def test_xspan_selector(backend: str):
    canvas = new_canvas(backend)
    mock = MagicMock()
    selector = tools.xspan_selector(canvas)
    selector.changed.connect(mock)
    canvas.mouse.emulate_drag([[0, 0], [1, 0]])
    assert mock.call_count == 1
    mock.assert_called_with((0, 1))

def test_yspan_selector(backend: str):
    canvas = new_canvas(backend)
    mock = MagicMock()
    selector = tools.yspan_selector(canvas)
    selector.changed.connect(mock)
    canvas.mouse.emulate_drag([[0, 0], [0, 1]])
    assert mock.call_count == 1
    mock.assert_called_with((0, 1))

def test_lasso_selector(backend: str):
    canvas = new_canvas(backend)
    mock = MagicMock()
    selector = tools.lasso_selector(canvas)
    selector.changed.connect(mock)
    canvas.mouse.emulate_drag([[0, 0], [0, 1], [1, 1], [1, 0]])
    assert mock.call_count == 1
    xy = np.array([(0, 0), (0, 1), (1, 1), (1, 0)])
    arg = mock.call_args[0][0]
    assert np.allclose(arg[0], xy[:, 0])
    assert np.allclose(arg[1], xy[:, 1])

def test_enabled():
    canvas = new_canvas("mock")
    mock = MagicMock()
    selector = tools.line_selector(canvas)
    selector.changed.connect(mock)
    assert selector.enabled
    mock.assert_not_called()
    canvas.mouse.emulate_drag([[0, 0], [0, 1], [1, 1]])
    mock.assert_called_once()
    mock.reset_mock()
    selector.enabled = False
    assert not selector.enabled
    canvas.mouse.emulate_drag([[0, 0], [0, 1], [1, 1]])
    mock.assert_not_called()

def test_tracking():
    canvas = new_canvas("mock")
    mock = MagicMock()
    selector = tools.line_selector(canvas, tracking=True)
    selector.changed.connect(mock)
    mock.assert_not_called()
    canvas.mouse.emulate_drag([[0, 0], [0, 1], [1, 1]])
    assert mock.call_count == 2

def test_point_contained():
    canvas = new_canvas("mock")
    with tools.rect_selector(canvas) as sel:
        canvas.mouse.emulate_drag([[0, 0], [1, 1]])
        assert sel.contains_point(0.5, 0.5)
        assert not sel.contains_point(1.5, 1.5)
        assert np.all(sel.contains_points([(0.2, 0.8), (1.3, 0.8)]) == [True, False])
    with tools.xspan_selector(canvas) as sel:
        canvas.mouse.emulate_drag([[0, 0], [1, 0]])
        assert sel.contains_point(0.5, 0)
        assert not sel.contains_point(1.5, 0)
        assert np.all(sel.contains_points([(0.2, 100), (1.3, -100)]) == [True, False])
    with tools.yspan_selector(canvas) as sel:
        canvas.mouse.emulate_drag([[0, 0], [0, 1]])
        assert sel.contains_point(0, 0.5)
        assert not sel.contains_point(0, 1.5)
        assert np.all(sel.contains_points([(100, 0.2), (-100, 1.3)]) == [True, False])
    with tools.lasso_selector(canvas) as sel:
        canvas.mouse.emulate_drag([[0, 0], [0, 1], [2, 1], [2, 2], [1, 2], [1, 0]])
        assert sel.contains_point([0.5, 0.5])
        assert not sel.contains_point(1.5, 0.5)
        assert np.all(sel.contains_points([(1.5, 1.4), (0.2, 1.8)]) == [True, False])



import numpy as np

from whitecanvas import new_canvas


def main():
    canvas = new_canvas(backend="matplotlib:qt")
    rng = np.random.default_rng(12345)
    x = np.arange(200)
    y = np.concatenate(
        [
            rng.normal(size=50),
            rng.normal(size=30, loc=2.3),
            rng.normal(size=50),
            rng.normal(size=40, loc=2.3),
            rng.normal(size=30),
        ],
    )
    canvas.add_line(x, y, width=1.5, color="black")
    canvas.add_spans([[50, 80], [130, 170]], color="red", alpha=0.3)
    canvas.show(block=True)

if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import Any, Callable, Protocol, runtime_checkable

import numpy as np
from cmap import Colormap
from numpy.typing import NDArray

from whitecanvas.types import Alignment, Hatch, LineStyle, MeshData, Orientation, Symbol

Array1D = NDArray[np.number]


@runtime_checkable
class BaseProtocol(Protocol):
    def _plt_get_visible(self) -> bool:
        """Return the visibility."""

    def _plt_set_visible(self, visible: bool):
        """Set the visibility."""


@runtime_checkable
class XYDataProtocol(BaseProtocol, Protocol):
    def _plt_get_data(self) -> tuple[Array1D, Array1D]:
        """Return the x and y array."""

    def _plt_set_data(self, xdata: Array1D, ydata: Array1D):
        """Set the x and y array."""


@runtime_checkable
class XYYDataProtocol(BaseProtocol, Protocol):
    def _plt_get_data(self) -> tuple[Array1D, Array1D, Array1D]:
        """Return the x and y array."""

    def _plt_set_data(self, xdata: Array1D, ydata0: Array1D, ydata1: Array1D):
        """Set the x and y array."""


@runtime_checkable
class OrientedXYYDataProtocol(BaseProtocol, Protocol):
    def _plt_get_vertical_data(self) -> tuple[Array1D, Array1D, Array1D]:
        """Return the vertical representative data."""

    def _plt_get_horizontal_data(self) -> tuple[Array1D, Array1D, Array1D]:
        """Return the horizontal representative data."""

    def _plt_set_vertical_data(self, xdata: Array1D, ydata0: Array1D, ydata1: Array1D):
        """Set the vertical representative data."""

    def _plt_set_horizontal_data(
        self, xdata: Array1D, ydata0: Array1D, ydata1: Array1D
    ):
        """Set the horizontal representative data."""


@runtime_checkable
class XXYYDataProtocol(BaseProtocol, Protocol):
    def _plt_get_data(self) -> tuple[Array1D, Array1D, Array1D, Array1D]:
        """Return the x and y array."""

    def _plt_set_data(
        self, xdata0: Array1D, xdata1: Array1D, ydata0: Array1D, ydata1: Array1D
    ):
        """Set the x and y array."""


@runtime_checkable
class SupportsMouseEvents(Protocol):
    def _plt_connect_pick_event(self, callback: Callable[[list[int]], Any]):
        """Connect the item picked event."""

    def _plt_set_hover_text(self, texts: list[str]):
        """Set hover texts to the lines"""


@runtime_checkable
class HasFaces(Protocol):
    def _plt_get_face_color(self) -> NDArray[np.float32]:
        """Return the face color."""

    def _plt_set_face_color(self, color: NDArray[np.float32]):
        """Set the face color."""

    def _plt_get_face_hatch(self) -> Hatch:
        """Return the face pattern."""

    def _plt_set_face_hatch(self, pattern: Hatch):
        """Set the face pattern."""


@runtime_checkable
class HasMultiFaces(Protocol):
    def _plt_get_face_color(self) -> NDArray[np.float32]:
        """Return the face color."""

    def _plt_set_face_color(self, color: NDArray[np.float32]):
        """Set the face color."""

    def _plt_get_face_hatch(self) -> list[Hatch]:
        """Return the face pattern."""

    def _plt_set_face_hatch(self, pattern: Hatch | list[Hatch]):
        """Set the face pattern."""


@runtime_checkable
class HasEdges(Protocol):
    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        """Return the edge color."""

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        """Set the edge color."""

    def _plt_get_edge_width(self) -> float:
        """Return the edge width."""

    def _plt_set_edge_width(self, width: float):
        """Set the edge width."""

    def _plt_get_edge_style(self) -> LineStyle:
        """Return the edge style."""

    def _plt_set_edge_style(self, style: LineStyle):
        """Set the edge style."""


@runtime_checkable
class HasMultiEdges(Protocol):
    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        """Return the edge color."""

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        """Set the edge color."""

    def _plt_get_edge_width(self) -> NDArray[np.floating]:
        """Return the edge width."""

    def _plt_set_edge_width(self, width: float | NDArray[np.floating]):
        """Set the edge width."""

    def _plt_get_edge_style(self) -> list[LineStyle]:
        """Return the edge style."""

    def _plt_set_edge_style(self, style: LineStyle | list[LineStyle]):
        """Set the edge style."""


@runtime_checkable
class HasText(BaseProtocol, Protocol):
    def _plt_get_text(self) -> list[str]:
        """Return the text."""

    def _plt_set_text(self, text: list[str]):
        """Set the text."""

    def _plt_get_text_color(self) -> NDArray[np.float32]:
        """Return the text color."""

    def _plt_set_text_color(self, color: NDArray[np.float32]):
        """Set the text color."""

    def _plt_get_text_size(self) -> NDArray[np.floating]:
        """Return the text size."""

    def _plt_set_text_size(self, size: NDArray[np.floating]):
        """Set the text size."""

    def _plt_get_text_position(
        self,
    ) -> tuple[NDArray[np.floating], NDArray[np.floating]]:
        """Return the text position."""

    def _plt_set_text_position(
        self, position: tuple[NDArray[np.floating], NDArray[np.floating]]
    ):
        """Set the text position."""

    def _plt_get_text_anchor(self) -> Alignment:
        """Return the text position."""

    def _plt_set_text_anchor(self, position: Alignment):
        """Set the text position."""

    def _plt_get_text_rotation(self) -> NDArray[np.floating]:
        """Return the text rotation in degree."""

    def _plt_set_text_rotation(self, rotation: NDArray[np.floating]):
        """Set the text rotation in degree."""

    def _plt_get_text_fontfamily(self) -> str:
        """Return the text font family."""

    def _plt_set_text_fontfamily(self, family: str):
        """Set the text font family."""


@runtime_checkable
class LineProtocol(XYDataProtocol, HasEdges, SupportsMouseEvents, Protocol):
    def _plt_get_antialias(self) -> bool:
        """Return the anti alias."""

    def _plt_set_antialias(self, antialias: bool):
        """Set the anti alias."""


@runtime_checkable
class MultiLineProtocol(XYDataProtocol, HasMultiEdges, SupportsMouseEvents, Protocol):
    def _plt_get_data(self) -> list[NDArray[np.number]]:
        """Return the x and y array."""

    def _plt_set_data(self, data: list[NDArray[np.number]]):
        """Set the x and y array."""

    def _plt_get_antialias(self) -> bool:
        """Return the anti alias."""

    def _plt_set_antialias(self, antialias: bool):
        """Set the anti alias."""


@runtime_checkable
class BarProtocol(
    XXYYDataProtocol, HasMultiFaces, HasMultiEdges, SupportsMouseEvents, Protocol
):
    """Protocols for plt.bar, plt.errorbar"""


@runtime_checkable
class MarkersProtocol(
    XYDataProtocol, HasMultiFaces, SupportsMouseEvents, HasMultiEdges, Protocol
):
    def _plt_get_symbol(self) -> Symbol:
        """Return the symbol."""

    def _plt_set_symbol(self, symbol: Symbol):
        """Set the symbol."""

    def _plt_get_symbol_size(self) -> NDArray[np.floating]:
        """Return the symbol size."""

    def _plt_set_symbol_size(self, size: float | NDArray[np.floating]):
        """Set the symbol size."""


@runtime_checkable
class BandProtocol(OrientedXYYDataProtocol, HasFaces, HasEdges, Protocol):
    def _plt_connect_pick_event(self, callback: Callable[[], Any]):
        """Connect the item picked event."""

    def _plt_set_hover_text(self, texts: str):
        """Set hover text to the band"""


@runtime_checkable
class ErrorbarProtocol(OrientedXYYDataProtocol, HasEdges, Protocol):
    def _plt_get_capsize(self) -> float:
        """Return the capsize of the error bar edges."""

    def _plt_set_capsize(self, capsize: float, orient: Orientation):
        """Set the capsize of the error bar edges."""


@runtime_checkable
class TextProtocol(HasText, HasEdges, HasFaces, Protocol):
    pass


@runtime_checkable
class VectorsProtocol(XXYYDataProtocol, HasEdges, Protocol):
    def _plt_get_antialias(self) -> bool:
        """Return the anti alias."""

    def _plt_set_antialias(self, antialias: bool):
        """Set the anti alias."""


@runtime_checkable
class MeshProtocol(BaseProtocol, HasFaces, HasEdges, Protocol):
    def _plt_get_data(self) -> MeshData:
        """Return the mesh data."""

    def _plt_set_data(self, data: MeshData):
        """Set the mesh data."""

    def _plt_get_face_color(self) -> NDArray[np.float32]:
        """Return the face color."""

    def _plt_set_face_color(self, color: NDArray[np.float32]):
        """Set the face color."""

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        """Return the edge color."""

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        """Set the edge color."""

    def _plt_get_edge_width(self) -> float:
        """Return the edge width."""

    def _plt_set_edge_width(self, width: float):
        """Set the edge width."""


@runtime_checkable
class ImageProtocol(BaseProtocol, Protocol):
    def _plt_get_data(self) -> NDArray[np.number]:
        """Return the image data."""

    def _plt_set_data(self, data: NDArray[np.number]):
        """Set the image data."""

    def _plt_get_colormap(self) -> Colormap:
        """Return the colormap."""

    def _plt_set_colormap(self, colormap: Colormap):
        """Set the colormap."""

    def _plt_get_clim(self) -> tuple[float, float]:
        """Return the clim."""

    def _plt_set_clim(self, clim: tuple[float, float]):
        """Set the clim."""

    def _plt_get_translation(self) -> tuple[float, float]:
        """Return the translation."""

    def _plt_set_translation(self, translation: tuple[float, float]):
        """Set the translation."""

    def _plt_get_scale(self) -> tuple[float, float]:
        """Return the scale."""

    def _plt_set_scale(self, scale: tuple[float, float]):
        """Set the scale."""


@runtime_checkable
class RangeDataProtocol(HasFaces, HasEdges, Protocol):
    def _plt_get_ranges(self) -> NDArray[np.number]:  # (N, 2)
        """Return the ranges."""

    def _plt_set_ranges(self, range: NDArray[np.number]):
        """Set the ranges."""



from __future__ import annotations

from typing import Iterable, Iterator, MutableSet, TypeVar

_V = TypeVar("_V")


class OrderedSet(MutableSet[_V]):
    """Implementation of an ordered set using a dict."""

    def __init__(self, values: Iterable[_V] = ()):
        self._dict = dict.fromkeys(values, None)  # NOTE: python dict is ordered

    def __contains__(self, key: _V):
        return key in self._dict

    def __iter__(self) -> Iterator[_V]:
        return iter(self._dict)

    def __len__(self) -> int:
        return len(self._dict)

    def add(self, key: _V):
        self._dict[key] = None

    def discard(self, key: _V):
        self._dict.pop(key, None)



from __future__ import annotations

from typing import TYPE_CHECKING, cast

import numpy as np
from vispy.scene import ArcballCamera, SceneCanvas, ViewBox, visuals
from vispy.util.quaternion import Quaternion

from whitecanvas import protocols
from whitecanvas.backend.vispy._label import TextLabel
from whitecanvas.backend.vispy.components3d.axis import Axis3D, AxisLabel3D

if TYPE_CHECKING:
    from vispy.scene import Grid


class Camera(ArcballCamera):
    def __init__(
        self,
        fov: float = 0,
        distance: None | float = None,
        translate_speed: float = 1,
        aspect_locked: bool = True,
        **kwargs,
    ):
        super().__init__(fov, distance, translate_speed, **kwargs)
        self._aspect_locked = aspect_locked


class Canvas3D:
    def __init__(self, viewbox: ViewBox):
        self._outer_viewbox = viewbox
        grid = cast("Grid", viewbox.add_grid())
        grid.spacing = 0
        camera = Camera()
        _viewbox = grid.add_view(row=1, col=0, camera=camera)
        self._viewbox: ViewBox = _viewbox

        title = TextLabel("")
        title.height_max = 40
        self._xaxis = Axis3D(self, 0)
        self._yaxis = Axis3D(self, 1)
        self._zaxis = Axis3D(self, 2)
        self._xlabel = AxisLabel3D(self._xaxis)
        self._ylabel = AxisLabel3D(self._yaxis)
        self._zlabel = AxisLabel3D(self._zaxis)
        grid.add_widget(title, row=0, col=0)
        self._title = title
        camera._quaternion = Quaternion(-0.4, 0.1, -0.9, 0.2)

    def _plt_add_layer(self, layer: visuals.visuals.Visual):
        layer.set_gl_state(
            depth_test=False,
            cull_face=False,
            blend=True,
            blend_func=("src_alpha", "one_minus_src_alpha", "one", "one"),
            blend_equation="func_add",
        )
        layer.parent = self._viewbox.scene

    def _plt_get_native(self):
        return self._viewbox

    def _plt_get_title(self):
        return self._title

    def _plt_reorder_layers(self, layers: list[protocols.BaseProtocol]):
        """Reorder layers in the canvas"""
        vb = self._viewbox
        for idx, layer in enumerate(layers):
            layer.order = idx
        if hasattr(vb, "_scene_ref"):
            scene: SceneCanvas = vb._scene_ref()
            scene._draw_order.clear()
            scene.update()

    def _plt_draw(self):
        pass

    def _plt_get_xaxis(self):
        return self._xaxis

    def _plt_get_yaxis(self):
        return self._yaxis

    def _plt_get_zaxis(self):
        return self._zaxis

    def _plt_get_xlabel(self):
        return self._xlabel

    def _plt_get_ylabel(self):
        return self._ylabel

    def _plt_get_zlabel(self):
        return self._zlabel

    def _plt_get_aspect_locked(self) -> bool:
        return True

    def _plt_set_aspect_locked(self, locked: bool):
        pass



from __future__ import annotations

import numpy as np
from cmap import Colormap
from numpy.typing import NDArray
from vispy.scene import visuals
from vispy.visuals.transforms import STTransform

from whitecanvas.protocols import ImageProtocol, check_protocol


@check_protocol(ImageProtocol)
class Image(visuals.Image):
    def __init__(self, data: np.ndarray):
        self._cmap_obj = Colormap("gray")
        # GPU does not support f64
        if data.dtype == np.float64:
            data = data.astype(np.float32)
        super().__init__(data, cmap="gray")
        tr = STTransform()
        self.transform = tr

    def _plt_get_visible(self) -> bool:
        return self.visible

    def _plt_set_visible(self, visible: bool):
        self.visible = visible

    def _plt_get_data(self) -> np.ndarray:
        return self._data

    def _plt_set_data(self, data: np.ndarray):
        self.set_data(data)

    def _plt_get_colormap(self) -> Colormap:
        return self._cmap_obj

    def _plt_set_colormap(self, cmap: Colormap):
        self.cmap = cmap.to_vispy()
        self._cmap_obj = cmap

    def _plt_get_clim(self) -> tuple[float, float]:
        return self.clim

    def _plt_set_clim(self, clim: tuple[float, float]):
        self.clim = clim

    def _plt_get_translation(self) -> NDArray[np.floating]:
        tr = self._plt_get_transform()
        dx, dy, *_ = tr.translate
        sx, sy = self._plt_get_scale()
        return (dx + 0.5 * sx, dy + 0.5 * sy)

    def _plt_set_translation(self, translation):
        tr = self._plt_get_transform()
        dx, dy = translation
        sx, sy = self._plt_get_scale()
        tr.translate = dx - 0.5 * sx, dy - 0.5 * sy, 0.0, 0.0
        self.transform = tr
        self.update()

    def _plt_get_scale(self) -> NDArray[np.floating]:
        tr = self._plt_get_transform()
        scale = tr.scale
        return scale[0], scale[1]

    def _plt_set_scale(self, scale):
        tr = self._plt_get_transform()
        dx, dy = self._plt_get_translation()
        sx, sy = scale[0], scale[1]
        tr.scale = (sx, sy, 1, 1)
        tr.translate = dx - 0.5 * sx, dy - 0.5 * sy, 0.0, 0.0
        self.transform = tr
        self.update()

    def _plt_get_transform(self) -> STTransform:
        tr = self.transform
        if not isinstance(tr, STTransform):
            raise TypeError(f"Unexpected transform type: {type(tr)}")
        return tr



from __future__ import annotations

import weakref

import numpy as np
import pyqtgraph as pg
from numpy.typing import NDArray
from pyqtgraph.GraphicsScene.mouseEvents import HoverEvent as pgHoverEvent
from qtpy import QtGui

from whitecanvas.backend.pyqtgraph._base import PyQtLayer
from whitecanvas.backend.pyqtgraph._qt_utils import (
    array_to_qcolor,
    from_qt_brush_style,
    from_qt_line_style,
    to_qt_brush_style,
    to_qt_line_style,
)
from whitecanvas.protocols import BandProtocol, check_protocol
from whitecanvas.types import Hatch, LineStyle, Orientation


@check_protocol(BandProtocol)
class Band(pg.FillBetweenItem, PyQtLayer):
    def __init__(self, t, ydata0, ydata1, orient: Orientation):
        if orient.is_vertical:
            c0 = pg.PlotCurveItem(t, ydata0)
            c1 = pg.PlotCurveItem(t, ydata1)
        else:
            c0 = pg.PlotCurveItem(ydata0, t)
            c1 = pg.PlotCurveItem(ydata1, t)
        pen = QtGui.QPen(QtGui.QColor(0, 0, 0))
        pen.setCosmetic(True)
        super().__init__(c0, c1, pen=pen, brush=QtGui.QBrush(QtGui.QColor(0, 0, 0)))
        self._viewBox = None
        self._hover_texts: list[str] = None
        self._callbacks = []

    def getViewBox(self):
        if self._viewBox is None:
            p = self
            while True:
                try:
                    p = p.parentItem()
                except RuntimeError:
                    return None
                if p is None:
                    vb = self.getViewWidget()
                    if vb is None:
                        return None
                    else:
                        self._viewBox = weakref.ref(vb)
                        break
                if hasattr(p, "implements") and p.implements("ViewBox"):
                    self._viewBox = weakref.ref(p)
                    break
        return self._viewBox()

    ##### XYDataProtocol #####
    def _plt_get_vertical_data(self):
        c0: pg.PlotCurveItem = self.curves[0]
        c1: pg.PlotCurveItem = self.curves[1]
        return c0.xData, c0.yData, c1.yData

    def _plt_set_vertical_data(self, t, ydata0, ydata1):
        c0: pg.PlotCurveItem = self.curves[0]
        c1: pg.PlotCurveItem = self.curves[1]
        c0.setData(t, ydata0)
        c1.setData(t, ydata1)

    def _plt_get_horizontal_data(self):
        c0: pg.PlotCurveItem = self.curves[0]
        c1: pg.PlotCurveItem = self.curves[1]
        return c0.yData, c0.xData, c1.xData

    def _plt_set_horizontal_data(self, y, xdata0, xdata1):
        c0: pg.PlotCurveItem = self.curves[0]
        c1: pg.PlotCurveItem = self.curves[1]
        c0.setData(xdata0, y)
        c1.setData(xdata1, y)

    ##### HasFace protocol #####
    def _get_brush(self) -> QtGui.QBrush:
        return self.brush()

    def _plt_get_face_color(self) -> NDArray[np.float32]:
        rgba = self._get_brush().color().getRgbF()
        return np.array(rgba)

    def _plt_set_face_color(self, color: NDArray[np.float32]):
        brush = self._get_brush()
        brush.setColor(array_to_qcolor(color))
        self.setBrush(brush)

    def _plt_get_face_hatch(self) -> Hatch:
        return from_qt_brush_style(self._get_brush().style())

    def _plt_set_face_hatch(self, pattern: Hatch):
        brush = self._get_brush()
        brush.setStyle(to_qt_brush_style(pattern))
        self.setBrush(brush)

    ##### HasEdges protocol #####
    def _get_pen(self) -> QtGui.QPen:
        return self.pen()

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        rgba = self._get_pen().color().getRgbF()
        return np.array(rgba)

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        pen = self._get_pen()
        pen.setColor(array_to_qcolor(color))
        self.setPen(pen)

    def _plt_get_edge_width(self) -> float:
        return self._get_pen().widthF()

    def _plt_set_edge_width(self, width: float):
        pen = self._get_pen()
        pen.setWidthF(width)
        self.setPen(pen)

    def _plt_get_edge_style(self) -> LineStyle:
        return from_qt_line_style(self._get_pen().style())

    def _plt_set_edge_style(self, style: LineStyle):
        pen = self._get_pen()
        pen.setStyle(to_qt_line_style(style))
        self.setPen(pen)

    def hoverEvent(self, ev: pgHoverEvent):
        vb = self.getViewBox()
        if vb is not None and self._hover_texts is not None:
            if self.contains(ev.pos()):
                self._toolTipCleared = False
                vb.setToolTip(self._hover_texts[0])
            else:
                if not self._toolTipCleared:
                    vb.setToolTip("")
                    self._toolTipCleared = True

    def _plt_set_hover_text(self, text: str):
        self._hover_texts = [text]

    def _plt_connect_pick_event(self, callback):
        def cb(ev: QtGui.QMouseEvent):
            if self.contains(ev.pos()):
                callback()

        self._callbacks.append(cb)

    def mousePressEvent(self, ev: QtGui.QMouseEvent):
        for cb in self._callbacks:
            cb(ev)
        return super().mousePressEvent(ev)



from __future__ import annotations

import bokeh.models as bk_models
import numpy as np
from cmap import Colormap

from whitecanvas.backend.bokeh._base import BokehLayer
from whitecanvas.protocols import ImageProtocol, check_protocol


@check_protocol(ImageProtocol)
class Image(BokehLayer[bk_models.Image]):
    def __init__(self, data: np.ndarray):
        self._data = bk_models.ColumnDataSource({"image": [data], "hovertexts": [""]})
        h, w = data.shape[:2]
        self._model = bk_models.Image(image="image", x=0, y=0, dw=w, dh=h)
        self._cmap = Colormap("gray")

    def _plt_get_data(self) -> np.ndarray:
        return self._data.data["image"][0]

    def _plt_set_data(self, data: np.ndarray):
        self._data.data = {"image": [data]}

    def _plt_get_colormap(self) -> Colormap:
        return self._cmap

    def _plt_set_colormap(self, cmap: Colormap):
        mapper = cmap.to_bokeh()
        mapper.low = self._model.color_mapper.low
        mapper.high = self._model.color_mapper.high
        self._model.color_mapper = mapper
        self._cmap = cmap

    def _plt_get_clim(self) -> tuple[float, float]:
        mapper = self._model.color_mapper
        return mapper.low, mapper.high

    def _plt_set_clim(self, clim: tuple[float, float]):
        self._model.color_mapper.low, self._model.color_mapper.high = clim

    def _plt_get_translation(self) -> tuple[float, float]:
        sx, sy = self._plt_get_scale()
        return self._model.x + 0.5 * sx, self._model.y + 0.5 * sy

    def _plt_set_translation(self, translation: tuple[float, float]):
        dx, dy = translation
        sx, sy = self._plt_get_scale()
        self._model.x, self._model.y = dx - 0.5 * sx, dy - 0.5 * sy

    def _plt_get_scale(self) -> tuple[float, float]:
        h, w = self._data.data["image"][0].shape[:2]
        return self._model.dw / w, self._model.dh / h

    def _plt_set_scale(self, scale: tuple[float, float]):
        h, w = self._data.data["image"][0].shape[:2]
        dx, dy = self._plt_get_translation()
        sx, sy = scale
        self._model.dw, self._model.dh = sx * w, sy * h
        self._model.x, self._model.y = dx - 0.5 * sx, dy - 0.5 * sy



from __future__ import annotations

from typing import Any, Sequence, TypeVar

import numpy as np

from whitecanvas.backend import Backend
from whitecanvas.layers._mixin import (
    EdgeNamespace,
    FaceNamespace,
    FontNamespace,
    TextMixin,
)
from whitecanvas.layers._sizehint import xy_size_hint
from whitecanvas.types import (
    Alignment,
    ArrayLike1D,
    ColorType,
    XYData,
    XYTextData,
    _Void,
)
from whitecanvas.utils.normalize import as_array_1d, normalize_xy

_Face = TypeVar("_Face", bound=FaceNamespace)
_Edge = TypeVar("_Edge", bound=EdgeNamespace)
_Font = TypeVar("_Font", bound=FontNamespace)
_void = _Void()


class Texts(TextMixin[_Face, _Edge, _Font]):
    def __init__(
        self,
        x: ArrayLike1D,
        y: ArrayLike1D,
        text: Sequence[str],
        *,
        name: str | None = None,
        color: ColorType = "black",
        size: float | None = None,
        rotation: float = 0.0,
        anchor: Alignment = Alignment.BOTTOM_LEFT,
        family: str | None = None,
        backend: Backend | str | None = None,
    ):
        super().__init__(name=name)
        x = as_array_1d(x)
        y = as_array_1d(y)
        self._backend = self._create_backend(Backend(backend), x, y, text)
        self.update(
            color=color,
            size=size,
            rotation=rotation,
            anchor=anchor,
            family=family,
        )
        pad = 0.0  # TODO: better padding
        self._x_hint, self._y_hint = xy_size_hint(x, y, pad, pad)

    def _get_layer_data(self) -> XYTextData:
        x, y = self._backend._plt_get_text_position()
        t = np.array(self._backend._plt_get_text(), dtype=np.object_)
        return XYTextData(x, y, t)

    def _norm_layer_data(self, data: Any) -> XYTextData:
        xpos, ypos, t = data
        if isinstance(t, str):
            t = [t] * self.ndata
        else:
            t = [str(t0) for t0 in t]
        xdata, ydata = normalize_xy(xpos, ypos)
        if len(xdata) != len(t):
            raise ValueError(
                f"Length of x ({len(xdata)}) and y ({len(ydata)}) must be equal "
                f"to the number of texts ({len(t)})."
            )
        return XYTextData(xdata, ydata, np.array(t, dtype=np.object_))

    def _set_layer_data(self, data: XYTextData):
        self._backend._plt_set_text_position((data.x, data.y))
        self._backend._plt_set_text(data.text.tolist())

    @property
    def ntexts(self):
        """Number of texts."""
        return self.ndata

    @property
    def ndata(self):
        """Number of texts."""
        return len(self.string)

    @property
    def string(self) -> list[str]:
        """Text strings."""
        return self._backend._plt_get_text()

    @string.setter
    def string(self, text: str | list[str]):
        if isinstance(text, str):
            text = [text] * self.ndata
        self._backend._plt_set_text(text)

    @property
    def pos(self) -> XYData:
        """Position of the text."""
        return XYData(*self._backend._plt_get_text_position())

    @pos.setter
    def pos(self, pos: tuple[ArrayLike1D, ArrayLike1D]):
        return self.set_pos(*pos)

    def set_pos(self, x: ArrayLike1D | None = None, y: ArrayLike1D | None = None):
        """Set the position of the text."""
        if x is None or y is None:
            x0, y0 = self.pos
            if x is None:
                x = x0
            if y is None:
                y = y0
        xdata, ydata = normalize_xy(x, y)
        if xdata.size != self.ndata:
            raise ValueError(
                f"Length of x ({xdata.size}) and y ({ydata.size}) must be equal "
                f"to the number of texts ({self.ndata})."
            )
        self._backend._plt_set_text_position((xdata, ydata))

    @property
    def color(self):
        """Color of the text."""
        return self.font.color

    @color.setter
    def color(self, color: ColorType | None):
        self.font.color = color

    @property
    def size(self):
        """Size of the text."""
        return self.font.size

    @size.setter
    def size(self, size: float | None):
        self.font.size = size

    @property
    def anchor(self) -> Alignment:
        """Anchor of the text."""
        return self._backend._plt_get_text_anchor()

    @anchor.setter
    def anchor(self, anc: str | Alignment):
        self._backend._plt_set_text_anchor(Alignment(anc))

    @property
    def rotation(self) -> float:
        """Rotation of the text."""
        return self._backend._plt_get_text_rotation()[0]

    @rotation.setter
    def rotation(self, rotation: float):
        self._backend._plt_set_text_rotation(np.full(self.ndata, float(rotation)))
        self.events.rotation.emit(rotation)

    @property
    def family(self) -> str:
        """Font family of the text."""
        return self.font.family

    @family.setter
    def family(self, fontfamily: str):
        self.font.family = fontfamily

    def update(
        self,
        *,
        color: ColorType | _Void = _void,
        size: float | _Void = _void,
        rotation: float | _Void = _void,
        anchor: Alignment | _Void = _void,
        family: str | _Void = _void,
    ) -> Texts:
        if rotation is not _void:
            self.rotation = rotation
        if anchor is not _void:
            self.anchor = anchor
        self.font.update(color=color, size=size, family=family)
        return self



import numpy as np

from whitecanvas import new_canvas


def rand(mean: float, n: int) -> list[float]:
    """Generate random data."""
    return np.random.normal(loc=mean, scale=mean / 4, size=n).tolist()

def main():
    # generate some random data
    np.random.seed(174623)
    data = {
        "label": ["Control"] * 50 + ["Treatment"] * 50,
        "value": rand(1.1, 15) + rand(1.4, 20) + rand(0.9, 15) + rand(3.3, 15) + rand(2.9, 20) + rand(3.8, 15),
        "replicate": [1] * 15 + [2] * 20 + [3] * 15 + [1] * 15 + [2] * 20 + [3] * 15,
    }

    canvas = new_canvas("matplotlib:qt", palette="colorbrewer:Pastel1")
    cat_plt = canvas.cat_x(data, x="label", y="value")

    # plot all the raw data
    cat_plt.add_swarmplot(color="replicate", size=8)

    # plot the mean of each replicate
    cat_plt.mean_for_each("replicate").add_markers(
        color="replicate", size=18, symbol="D"
    )

    # plot the mean of all the data for control and treatment
    cat_plt.mean().add_markers(color="black", size=20, symbol="+").with_edge(width=3)

    # plot the mean of replicate means
    cat_plt.mean_for_each("replicate").mean().add_markers(
        color="black", size=30, symbol="_"
    )
    canvas.show(block=True)

if __name__ == "__main__":
    main()



"""Built-in tools."""

from whitecanvas.tools._selection import (
    lasso_selector,
    line_selector,
    polygon_selector,
    rect_selector,
    xspan_selector,
    yspan_selector,
)

__all__ = [
    "line_selector",
    "rect_selector",
    "xspan_selector",
    "yspan_selector",
    "polygon_selector",
    "lasso_selector",
]



from __future__ import annotations

from typing import Any, Iterable, Literal

import numpy as np
from numpy.typing import ArrayLike

from whitecanvas.backend import Backend
from whitecanvas.layers._mixin import CollectionFaceEdgeMixin
from whitecanvas.layers._primitive import Band
from whitecanvas.layers.group._cat_utils import check_array_input
from whitecanvas.layers.group._collections import (
    LayerCollectionBase,
    RichContainerEvents,
)
from whitecanvas.types import Orientation, XYYData
from whitecanvas.utils.normalize import as_array_1d, parse_texts


class BandCollection(
    LayerCollectionBase[Band],
    CollectionFaceEdgeMixin,
):
    events: RichContainerEvents
    _events_class = RichContainerEvents

    def __init__(
        self,
        data: list[XYYData],
        *,
        orient: Orientation = Orientation.VERTICAL,
        name: str | None = None,
        backend: str | Backend | None = None,
    ):
        bands = [Band(*each, orient=orient, backend=backend) for each in data]
        LayerCollectionBase.__init__(self, bands, name=name)
        CollectionFaceEdgeMixin.__init__(self)
        self._orient = orient
        self._init_events()

    @property
    def data(self) -> list[XYYData]:
        return [line.data for line in self]

    @data.setter
    def data(self, data: list[XYYData]):
        ndata_in = len(data)
        ndata_now = len(self)
        if ndata_in > ndata_now:
            for _ in range(ndata_now, ndata_in):
                self.append(Band([], [], []))
        elif ndata_in < ndata_now:
            for _ in range(ndata_in, ndata_now):
                del self[-1]
        for line, d in zip(self, data):
            line.data = d

    @property
    def orient(self) -> Orientation:
        """Orientation of the bands."""
        return self._orient

    def with_hover_text(self, text: str | Iterable[Any]):
        """Set hover text for each band."""
        if isinstance(text, str):
            texts = [text] * len(self)
        else:
            texts = [str(t) for t in text]
        for band, t in zip(self, texts):
            band.with_hover_text(t)
        return self

    def with_hover_template(self, template: str, extra: dict[str, Any] | None = None):
        if self._backend_name in ("plotly", "bokeh"):  # conversion for HTML
            template = template.replace("\n", "<br>")
        params = parse_texts(template, len(self), extra)
        # set default format keys
        if "i" not in params:
            params["i"] = np.arange(len(self))
        texts = [
            template.format(**{k: v[i] for k, v in params.items()})
            for i in range(len(self))
        ]
        return self.with_hover_text(texts)


class ViolinPlot(BandCollection):
    def __init__(
        self,
        data: list[XYYData],
        *,
        name: str | None = None,
        shape: Literal["both", "left", "right"] = "both",
        extent: float = 0.5,
        orient: Orientation = Orientation.VERTICAL,
        backend: str | Backend | None = None,
    ):
        super().__init__(data, name=name, orient=orient, backend=backend)
        self._shape = shape
        self._extent = extent

    @classmethod
    def from_arrays(
        cls,
        x: list[float],
        data: list[ArrayLike],
        *,
        name: str | None = None,
        orient: str | Orientation = Orientation.VERTICAL,
        shape: Literal["both", "left", "right"] = "both",
        extent: float = 0.5,
        kde_band_width: float | str = "scott",
        backend: str | Backend | None = None,
    ):
        ori = Orientation.parse(orient)
        new_vals = cls._convert_data(
            x, data, shape=shape, extent=extent, kde_band_width=kde_band_width
        )
        return cls(
            new_vals,
            name=name,
            shape=shape,
            extent=extent,
            orient=ori.transpose(),
            backend=backend,
        )

    @property
    def orient(self) -> Orientation:
        """Orientation of the violin plot (perpendicular to the fill orientation)."""
        return self._orient.transpose()

    @property
    def ndata(self) -> int:
        return len(self)

    @property
    def extent(self):
        return self._extent

    @extent.setter
    def extent(self, width: float):
        if width <= 0:
            raise ValueError(f"extent must be positive, got {width}")
        factor = width / self.extent
        for band in self:
            bd = band.data
            ycenter = bd.ycenter
            y0 = (bd.y0 - ycenter) * factor + ycenter
            y1 = (bd.y1 - ycenter) * factor + ycenter
            band.set_data(bd.x, y0, y1)
        self._extent = width

    @staticmethod
    def _convert_data(
        x: list[float],
        data: list[ArrayLike],
        shape: Literal["both", "left", "right"] = "both",
        extent: float = 0.5,
        kde_band_width: float | str = "scott",
    ):
        from whitecanvas.utils.kde import gaussian_kde

        if extent <= 0:
            raise ValueError(f"extent must be positive, got {extent}")
        x, data = check_array_input(x, data)
        xyy_values: list[XYYData] = []
        for offset, values in zip(x, data):
            arr = as_array_1d(values)
            kde = gaussian_kde(arr, bw_method=kde_band_width)

            sigma = np.sqrt(kde.covariance[0, 0])
            pad = sigma * 2.5
            x_ = np.linspace(arr.min() - pad, arr.max() + pad, 100)
            y = kde(x_)
            if shape in ("both", "left"):
                y0 = -y + offset
            else:
                y0 = np.zeros_like(y) + offset
            if shape in ("both", "right"):
                y1 = y + offset
            else:
                y1 = np.zeros_like(y) + offset

            data = XYYData(x_, y0, y1)
            xyy_values.append(data)

        half_widths: list[float] = []
        for xyy in xyy_values:
            half_width = np.max(np.abs(xyy.ydiff))
            if shape == "both":
                half_width /= 2
            half_widths.append(half_width)
        factor = extent / np.max(half_widths) / 2
        new_vals: list[XYYData] = []
        for xyy, xoffset in zip(xyy_values, x):
            y0 = (xyy.y0 - xoffset) * factor + xoffset
            y1 = (xyy.y1 - xoffset) * factor + xoffset
            new_vals.append(XYYData(xyy.x, y0, y1))
        return new_vals



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray

# modified from: https://qiita.com/yamadasuzaku/items/b7482131a06759731b47


def is_in_polygon(points: NDArray[np.number], poly: NDArray[np.number]):
    num = poly.shape[0]
    inside = np.full(points.shape[0], False, dtype=np.bool_)
    xs = points[:, 0]
    ys = points[:, 1]
    p1x, p1y = poly[0]
    for i in range(1, num + 1):
        p2x, p2y = poly[i % num]
        mask = (
            (ys > min(p1y, p2y))
            & (ys <= max(p1y, p2y))
            & (xs <= max(p1x, p2x))
            & (p1y != p2y)
        )
        xinters = (ys[mask] - p1y) * (p2x - p1x) / (p2y - p1y) + p1x
        inside[mask] = np.logical_xor(inside[mask], xs[mask] < xinters)
        p1x, p1y = p2x, p2y

    return inside



from __future__ import annotations

import numpy as np
from mpl_toolkits.mplot3d import art3d
from numpy.typing import NDArray

from whitecanvas.backend.matplotlib._base import MplLayer
from whitecanvas.types import LineStyle


class MonoLine3D(art3d.Line3D, MplLayer):
    def __init__(self, xdata, ydata, zdata):
        super().__init__(
            xdata,
            ydata,
            zdata,
            linewidth=1,
            linestyle="-",
            color="blue",
            markersize=0,
            # picker=True,
            # pickradius=5,
        )

    ##### XYDataProtocol #####
    def _plt_get_data(self):
        return self.get_data_3d()

    def _plt_set_data(self, xdata, ydata, zdata):
        self.set_data_3d(xdata, ydata, zdata)

    ##### HasEdges #####
    def _plt_get_edge_width(self) -> float:
        return self.get_linewidth()

    def _plt_set_edge_width(self, width: float):
        self.set_linewidth(width)

    def _plt_get_edge_style(self) -> LineStyle:
        return LineStyle(self.get_linestyle())

    def _plt_set_edge_style(self, style: LineStyle):
        self.set_linestyle(style.value)

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return self.get_color()

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        self.set_color(color)

    def _plt_get_antialias(self) -> bool:
        return self.get_antialiased()

    def _plt_set_antialias(self, antialias: bool):
        self.set_antialiased(antialias)



import numpy as np
from numpy.testing import assert_allclose, assert_array_equal

from whitecanvas import new_canvas


def test_line_marker_text(backend: str):
    canvas = new_canvas(backend=backend)
    layer = canvas.add_line(
        np.arange(10), np.arange(10) * 2
    ).with_markers(
        symbol="s", size=10, color="red"
    ).with_text(
        [f"Data-{i}" for i in range(10)], color="blue"
    ).with_yerr(
        np.ones(10), color="blue"
    )
    assert_array_equal(layer.data.x, np.arange(10))
    assert_array_equal(layer.data.y, np.arange(10) * 2)
    assert len(canvas.layers) == 1


def test_marker_err(backend: str):
    canvas = new_canvas(backend=backend)
    rng = np.random.default_rng(0)
    data = rng.random((10, 2))
    layer = canvas.add_markers(data).with_xerr(np.ones(10), color="blue")
    layer = canvas.add_markers(data).with_face(color="red").with_yerr(np.ones(10), color="blue")

    assert_allclose(layer.data.x, data[:, 0], atol=1e-6)
    assert_allclose(layer.data.y, data[:, 1], atol=1e-6)

    canvas.add_markers(rng.normal(size=(50, 2))).color_by_density()

def test_bar_err_text(backend: str):
    canvas = new_canvas(backend=backend)
    layer = canvas.add_bars(
        np.arange(10), np.arange(10) * 2
    ).with_err(
        np.ones(10), color="blue"
    ).with_text(
        [f"Data-{i}" for i in range(10)], color="blue"
    )
    assert_array_equal(layer.data.x, np.arange(10))
    assert_allclose(layer.data.y, np.arange(10) * 2, atol=1e-6)
    assert len(canvas.layers) == 1

def test_stem(backend: str):
    canvas = new_canvas(backend=backend)
    x, y = np.arange(10), np.arange(10) * 2
    layer = canvas.add_markers(x, y).with_stem(bottom=0.1)
    layer = canvas.add_markers(x, y).with_stem(orient="horizontal")
    assert_array_equal(layer.data.x, np.arange(10))
    assert_allclose(layer.data.y, np.arange(10) * 2, atol=1e-6)
    assert len(canvas.layers) == 2

def test_network(backend: str):
    canvas = new_canvas(backend=backend)
    layer = (
        canvas.add_markers(np.arange(10), np.arange(10) * 2)
        .with_network([[0, 1], [0, 2], [1, 3]])
        .with_text_offset(0.1, 0.1)
    )
    assert_array_equal(layer.nodes.data.x, np.arange(10))
    assert_array_equal(layer.nodes.data.y, np.arange(10) * 2)
    assert len(canvas.layers) == 1

def test_line_fill(backend: str):
    canvas = new_canvas(backend=backend)
    canvas.add_line([-1.5, -1, 0, 1, 1.5]).with_yfill()
    canvas.add_line([-1.5, -1, 0, 1, 1.5], [0, 0, 0, 0, 0]).with_xfill()

def test_image(backend: str):
    canvas = new_canvas(backend=backend)
    rng = np.random.default_rng(0)
    img = canvas.add_image(rng.random((3, 3)))
    img = img.with_text()
    img = img.with_colorbar()
    mask = img.data > 0.5
    img.with_text(mask=mask)



from __future__ import annotations

from typing import TYPE_CHECKING, overload

from whitecanvas.canvas import Canvas
from whitecanvas.plot._canvases import current_canvas


def _make_method(name: str, pref: str = "add_"):
    fname = f"{pref}{name}"

    def _inner(*args, **kwargs):
        meth = getattr(current_canvas(), fname)
        return meth(*args, **kwargs)

    canvas_func = getattr(Canvas, fname)
    doc = canvas_func.__doc__
    assert isinstance(doc, str)
    _inner.__doc__ = doc.replace(f">>> canvas.{pref}", ">>> plt.")
    _inner.__name__ = name
    _inner.__qualname__ = name
    _inner.__module__ = "whitecanvas.plot._methods"
    _inner.__annotations__ = canvas_func.__annotations__
    return _inner


if TYPE_CHECKING:
    _CANVAS = Canvas(backend="mock")
    # add-layer methods
    line = _CANVAS.add_line
    step = _CANVAS.add_step
    hline = _CANVAS.add_hline
    vline = _CANVAS.add_vline
    markers = _CANVAS.add_markers
    hist = _CANVAS.add_hist
    bars = _CANVAS.add_bars
    band = _CANVAS.add_band
    errorbars = _CANVAS.add_errorbars
    infline = _CANVAS.add_infline
    infcurve = _CANVAS.add_infcurve
    spans = _CANVAS.add_spans
    kde = _CANVAS.add_kde
    rug = _CANVAS.add_rug
    text = _CANVAS.add_text
    vectors = _CANVAS.add_vectors
    # categorical methods
    cat = _CANVAS.cat
    cat_x = _CANVAS.cat_x
    cat_y = _CANVAS.cat_y
    cat_xy = _CANVAS.cat_xy
    # update methods
    update_axes = _CANVAS.update_axes
    update_labels = _CANVAS.update_labels
    update_font = _CANVAS.update_font
    # others
    legend = _CANVAS.add_legend
else:
    line = _make_method("line")
    step = _make_method("step")
    hline = _make_method("hline")
    vline = _make_method("vline")
    markers = _make_method("markers")
    hist = _make_method("hist")
    bars = _make_method("bars")
    band = _make_method("band")
    errorbars = _make_method("errorbars")
    infline = _make_method("infline")
    infcurve = _make_method("infcurve")
    spans = _make_method("spans")
    kde = _make_method("kde")
    rug = _make_method("rug")
    text = _make_method("text")
    vectors = _make_method("vectors")
    cat = _make_method("cat", pref="")
    cat_x = _make_method("cat_x", pref="")
    cat_y = _make_method("cat_y", pref="")
    cat_xy = _make_method("cat_xy", pref="")
    update_axes = _make_method("update_axes", pref="")
    update_labels = _make_method("update_labels", pref="")
    update_font = _make_method("update_font", pref="")
    legend = _make_method("legend", pref="add_")


@overload
def xlim() -> tuple[float, float]: ...


@overload
def xlim(min: float, max: float, /) -> tuple[float, float]: ...


@overload
def xlim(minmax: tuple[float, float], /) -> tuple[float, float]: ...


def xlim(*args):
    if len(args) == 0:
        return current_canvas().x.lim
    elif len(args) == 1:
        x0, x1 = args[0]
    else:
        x0, x1 = args
    current_canvas().x.lim = x0, x1
    return current_canvas().x.lim


@overload
def ylim() -> tuple[float, float]: ...


@overload
def ylim(min: float, max: float, /) -> tuple[float, float]: ...


@overload
def ylim(minmax: tuple[float, float], /) -> tuple[float, float]: ...


def ylim(*args):
    if len(args) == 0:
        return current_canvas().y.lim
    elif len(args) == 1:
        y0, y1 = args[0]
    else:
        y0, y1 = args
    current_canvas().y.lim = y0, y1
    return current_canvas().y.lim


def title(title: str | None = None) -> str:
    if title is not None:
        current_canvas().title = title
    return current_canvas().title


def xlabel(label: str | None = None) -> str:
    if label is not None:
        current_canvas().x.label = label
    return current_canvas().x.label


def ylabel(label: str | None = None) -> str:
    if label is not None:
        current_canvas().y.label = label
    return current_canvas().y.label


def xticks(ticks: list[float] | None = None) -> list[float]:
    if ticks is not None:
        current_canvas().x.ticks = ticks
    return current_canvas().x.ticks


def yticks(ticks: list[float] | None = None) -> list[float]:
    if ticks is not None:
        current_canvas().y.ticks = ticks
    return current_canvas().y.ticks



from __future__ import annotations

from typing import Generic, TypeVar

import numpy as np
from numpy.typing import NDArray

from whitecanvas.backend import Backend
from whitecanvas.layers import _legend
from whitecanvas.layers._mixin import EdgeNamespace, FaceNamespace, MonoEdge, MonoFace
from whitecanvas.layers._primitive import Markers, MultiLine
from whitecanvas.layers.group._collections import LayerContainer
from whitecanvas.types import ArrayLike1D, Orientation, XYData
from whitecanvas.utils.normalize import normalize_xy

_Face = TypeVar("_Face", bound=FaceNamespace)
_Edge = TypeVar("_Edge", bound=EdgeNamespace)
_Size = TypeVar("_Size", float, NDArray[np.floating])


class StemPlot(LayerContainer, Generic[_Face, _Edge, _Size]):
    _ATTACH_TO_AXIS = True

    def __init__(
        self,
        markers: Markers[_Face, _Edge, _Size],
        lines: MultiLine,
        *,
        name: str | None = None,
        orient: Orientation = Orientation.VERTICAL,
    ):
        self._orient = Orientation.parse(orient)
        super().__init__([markers, lines], name=name)

    def _default_ordering(self, n: int) -> list[int]:
        assert n == 2
        return [1, 0]

    @property
    def data(self) -> XYData:
        """XYData as (x, height)."""
        mdata = self.markers.data
        if self.orient.is_vertical:
            xdata = mdata.x
            ydata = mdata.y - self.bottom
        else:
            xdata = mdata.x - self.bottom
            ydata = mdata.y
        return XYData(xdata, ydata)

    @property
    def bottom(self) -> NDArray[np.floating]:
        """Bottom of the stem."""
        if self.orient.is_vertical:
            return np.array([d[0, 1] for d in self.lines.data])
        else:
            return np.array([d[0, 0] for d in self.lines.data])

    @property
    def top(self) -> NDArray[np.floating]:
        """Top of the stem."""
        if self.orient.is_vertical:
            return self.markers.data.y
        else:
            return self.markers.data.x

    @property
    def markers(self) -> Markers[_Face, _Edge, _Size]:
        """Markers layer."""
        return self._children[0]

    @property
    def lines(self) -> MultiLine:
        """Lines for the stems."""
        return self._children[1]

    @property
    def orient(self) -> Orientation:
        """Orientation of the stem plot."""
        return self._orient

    def bbox_hint(self) -> NDArray[np.float64]:
        xmin, xmax, ymin, ymax = self.markers.bbox_hint()
        # NOTE: min/max is nan-safe (returns nan)
        if self.orient.is_vertical:
            ymin = min(ymin, 0)
            ymax = max(ymax, 0)
        else:
            xmin = min(xmin, 0)
            xmax = max(xmax, 0)
        return xmin, xmax, ymin, ymax

    @classmethod
    def from_arrays(
        cls,
        xdata: ArrayLike1D,
        ydata: ArrayLike1D,
        bottom: ArrayLike1D | float = 0,
        name: str | None = None,
        orient: Orientation = Orientation.VERTICAL,
        backend: str | Backend | None = None,
    ) -> StemPlot[MonoFace, MonoEdge, float]:
        xdata, ydata = normalize_xy(xdata, ydata)
        top = ydata + bottom
        return Markers(xdata, top, name=name, backend=backend).with_stem(
            orient, bottom=bottom
        )

    def _as_legend_item(self) -> _legend.StemLegendItem:
        return _legend.StemLegendItem(
            self.lines._as_legend_item(), self.markers._as_legend_item()
        )



from __future__ import annotations

from typing import Callable, Sequence, TypeVar

_T = TypeVar("_T")


def make_sorter_manual(order: Sequence[_T]) -> Callable[[Sequence[_T]], Sequence[_T]]:
    if len(order) != len(set(order)):
        raise ValueError(f"Order contains duplicate values: {order!r}")

    def _sort_manual(x: Sequence[_T]) -> Sequence[_T]:
        out: list[_T] = []
        input_categories = set(x)
        for each in order:
            if each in input_categories:
                out.append(each)
                input_categories.discard(each)
        if input_categories:
            raise ValueError(
                f"Categories {input_categories!r} are not in the given order "
                f"{order!r}"
            )
        return order

    return _sort_manual


def sorter_ascending(x: Sequence[_T]) -> Sequence[_T]:
    return sorted(x)


def sorter_descending(x: Sequence[_T]) -> Sequence[_T]:
    return sorted(x, reverse=True)



from __future__ import annotations

import numpy as np
from cmap import Colormap
from plotly import graph_objects as go

from whitecanvas.backend.plotly._base import PlotlyLayer
from whitecanvas.protocols import ImageProtocol, check_protocol


@check_protocol(ImageProtocol)
class Image(PlotlyLayer[go.Heatmap]):
    def __init__(self, data: np.ndarray):
        self._props = {
            "z": data,
            "visible": True,
            "type": "heatmap",
            "colorscale": "gray",
            "zmin": np.min(data),
            "zmax": np.max(data),
            "dx": 1,
            "dy": 1,
            "x0": 0,
            "y0": 0,
            "showlegend": False,
        }
        self._cmap = Colormap("gray")

    def _plt_get_data(self) -> np.ndarray:
        return self._props["z"]

    def _plt_set_data(self, data: np.ndarray):
        self._props["z"] = data

    def _plt_get_colormap(self) -> Colormap:
        return self._cmap

    def _plt_set_colormap(self, cmap: Colormap):
        self._props["colorscale"] = cmap.to_plotly()
        self._cmap = cmap

    def _plt_get_clim(self) -> tuple[float, float]:
        return self._props["zmin"], self._props["zmax"]

    def _plt_set_clim(self, clim: tuple[float, float]):
        self._props["zmin"], self._props["zmax"] = clim

    def _plt_get_visible(self) -> bool:
        return self._props["visible"]

    def _plt_get_translation(self) -> tuple[float, float]:
        return self._props["x0"], self._props["y0"]

    def _plt_set_translation(self, translation: tuple[float, float]):
        self._props["x0"], self._props["y0"] = translation

    def _plt_get_scale(self) -> tuple[float, float]:
        return self._props["dx"], self._props["dy"]

    def _plt_set_scale(self, scale: tuple[float, float]):
        self._props["dx"], self._props["dy"] = scale



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, ClassVar
from weakref import WeakSet

if TYPE_CHECKING:
    from whitecanvas.canvas._namespaces import AxisNamespace


class AxisLinker:
    _GLOBAL_LINKERS: ClassVar[set[AxisLinker]] = set()

    def __new__(cls):
        self = super().__new__(cls)
        cls._GLOBAL_LINKERS.add(self)
        return self

    def __init__(self):
        self._axis_set = WeakSet["AxisNamespace"]()
        self._updating = False

    def link(self, axis: AxisNamespace):
        """Link an axis."""
        axis._get_canvas()  # raise error if the parent canvas is deleted.
        if axis in self._axis_set:
            warnings.warn(f"Axis {axis} already linked", RuntimeWarning, stacklevel=2)
            return
        self._axis_set.add(axis)
        axis.events.lim.connect(self.set_limits, max_args=1)

    def unlink(self, axis: AxisNamespace):
        """Unlink an axis."""
        if axis not in self._axis_set:
            warnings.warn(f"Axis {axis} was not linked", RuntimeWarning, stacklevel=2)
        self._axis_set.discard(axis)
        axis.events.lim.disconnect(self.set_limits)

    def unlink_all(self) -> None:
        """Unlink all axes."""
        axes = list(self._axis_set)  # avoid the size changing during iteration
        for axis in axes:
            self.unlink(axis)
        self.__class__._GLOBAL_LINKERS.discard(self)

    def is_alive(self) -> bool:
        """Check if the linker is still alive."""
        return self in self.__class__._GLOBAL_LINKERS

    def set_limits(self, limits: tuple[float, float]):
        if self._updating:
            return
        self._updating = True
        try:
            for axis in self._axis_set:
                axis.lim = limits
        finally:
            self._updating = False

    @classmethod
    def link_axes(cls, *axes: AxisNamespace):
        """Link multiple axes."""
        self = cls()
        if len(axes) == 1 and hasattr(axes[0], "__iter__"):
            axes = axes[0]
        for axis in axes:
            self.link(axis)
        return self


class AxisLinkerRef:
    def __init__(self, linker: AxisLinker):
        self._linker = linker

    def _get_linker(self):
        if self._linker.is_alive():
            return self._linker
        raise RuntimeError("Linker has been deleted")

    def link(self, axis: AxisNamespace):
        """Link an axis."""
        self._get_linker().link(axis)
        return self

    def unlink(self, axis: AxisNamespace):
        """Unlink an axis."""
        self._get_linker().unlink(axis)
        return self

    def unlink_all(self) -> None:
        """Unlink all axes."""
        self._get_linker().unlink_all()


def link_axes(*axes: AxisNamespace):
    """Link multiple axes."""
    linker = AxisLinker.link_axes(*axes)
    return AxisLinkerRef(linker)



from __future__ import annotations

from typing import Sequence

import numpy as np
from numpy.typing import NDArray

from whitecanvas.types import Hatch, LineStyle
from whitecanvas.utils.normalize import as_color_array
from whitecanvas.utils.type_check import is_real_number


class BaseMockLayer:
    def __init__(self):
        self._visible = True

    def _plt_get_visible(self):
        return self._visible

    def _plt_set_visible(self, visible):
        self._visible = visible


class MockHasData(BaseMockLayer):
    def __init__(self, data):
        super().__init__()
        self._data = data

    def _plt_get_data(self) -> Sequence[NDArray[np.float32]]:
        return self._data

    def _plt_set_data(self, *data: Sequence[NDArray[np.float32]]):
        self._data = data


class MockHasFaces:
    def _plt_get_face_color(self) -> NDArray[np.float32]:
        if not hasattr(self, "_face_color"):
            self._face_color = np.zeros(4, dtype=np.float32)
        return self._face_color

    def _plt_set_face_color(self, color: NDArray[np.float32]):
        self._face_color = color

    def _plt_get_face_hatch(self) -> Hatch:
        if not hasattr(self, "_face_hatch"):
            self._face_hatch = Hatch.SOLID
        return self._face_hatch

    def _plt_set_face_hatch(self, pattern: Hatch):
        self._face_hatch = pattern


class MockHasEdges:
    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        if not hasattr(self, "_edge_color"):
            self._edge_color = np.zeros(4, dtype=np.float32)
        return self._edge_color

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        self._edge_color = color

    def _plt_get_edge_style(self) -> LineStyle:
        if not hasattr(self, "_edge_style"):
            self._edge_style = LineStyle.SOLID
        return self._edge_style

    def _plt_set_edge_style(self, style: LineStyle):
        self._edge_style = style

    def _plt_get_edge_width(self) -> float:
        if not hasattr(self, "_edge_width"):
            self._edge_width = 1.0
        return self._edge_width

    def _plt_set_edge_width(self, width: float):
        self._edge_width = width


class MockHasMultiFaces:
    def _plt_get_ndata(self) -> int:
        raise NotImplementedError

    def _plt_get_face_color(self) -> NDArray[np.float32]:
        if not hasattr(self, "_face_color"):
            ndata = self._plt_get_ndata()
            self._face_color = np.zeros((ndata, 4), dtype=np.float32)
        return self._face_color

    def _plt_set_face_color(self, color: NDArray[np.float32]):
        self._face_color = as_color_array(color, self._plt_get_ndata())

    def _plt_get_face_hatch(self) -> list[Hatch]:
        if not hasattr(self, "_face_hatch"):
            self._face_hatch = [Hatch.SOLID] * self._plt_get_ndata()
        return self._face_hatch

    def _plt_set_face_hatch(self, pattern: Hatch | Sequence[Hatch]):
        if isinstance(pattern, (str, Hatch)):
            pattern = [Hatch(pattern)] * self._plt_get_ndata()
        else:
            pattern = [Hatch(p) for p in pattern]
        self._face_hatch = pattern


class MockHasMultiEdges:
    def _plt_get_ndata(self) -> int:
        raise NotImplementedError

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        if not hasattr(self, "_edge_color"):
            ndata = self._plt_get_ndata()
            self._edge_color = np.zeros((ndata, 4), dtype=np.float32)
        return self._edge_color

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        self._edge_color = as_color_array(color, self._plt_get_ndata())

    def _plt_get_edge_style(self) -> list[LineStyle]:
        if not hasattr(self, "_edge_style"):
            self._edge_style = [LineStyle.SOLID] * self._plt_get_ndata()
        return self._edge_style

    def _plt_set_edge_style(self, style: LineStyle | Sequence[LineStyle]):
        if isinstance(style, (str, LineStyle)):
            style = [LineStyle(style)] * self._plt_get_ndata()
        else:
            style = [LineStyle(s) for s in style]
        self._edge_style = style

    def _plt_get_edge_width(self) -> float:
        if not hasattr(self, "_edge_width"):
            self._edge_width = np.ones(self._plt_get_ndata(), dtype=np.float32)
        return self._edge_width

    def _plt_set_edge_width(self, width: float | Sequence[float]):
        if is_real_number(width):
            width = np.full(self._plt_get_ndata(), width, dtype=np.float32)
        else:
            width = np.asarray(width)
        self._edge_width = width


class MockHasMouseEvents:
    def _plt_set_hover_text(self, text: list[str]) -> None:
        pass

    def _plt_connect_pick_event(self, callback) -> None:
        pass



from __future__ import annotations

from typing import TYPE_CHECKING, Generic, TypeVar

from numpy.typing import ArrayLike

from whitecanvas.layers import _legend, _text_utils
from whitecanvas.layers._primitive import Errorbars, Markers, Texts
from whitecanvas.layers._primitive.line import _SingleLine
from whitecanvas.layers.group._collections import LayerContainer
from whitecanvas.types import Alignment, ColorType, LineStyle, Orientation, _Void

if TYPE_CHECKING:
    from whitecanvas.layers._primitive.markers import _Edge, _Face, _Size
    from whitecanvas.layers.group.labeled import LabeledPlot

_void = _Void()
_L = TypeVar("_L", bound=_SingleLine)


class Plot(LayerContainer, Generic[_L]):
    """A Plot layer is composed of a line and a markers layer."""

    def __init__(
        self,
        line: _L,
        markers: Markers[_Face, _Edge, _Size],
        name: str | None = None,
    ):
        super().__init__([line, markers], name=name)

    @property
    def line(self) -> _L:
        """The line layer."""
        return self._children[0]

    @property
    def markers(self) -> Markers[_Face, _Edge, _Size]:
        """The markers layer."""
        return self._children[1]

    def with_edge(
        self,
        color: ColorType | None = None,
        alpha: float = 1.0,
        width: float = 1.0,
        style: str | LineStyle = LineStyle.SOLID,
    ) -> Plot:
        """
        Set the edge properties of the markers.

        Parameters
        ----------
        color : ColorType, optional
            The color of the edge. If not given, use the color defined by the theme.
        alpha : float, default 1.0
            The transparency of the edge.
        width : float, default 1.0
            The width of the edge.
        style : str or LineStyle, default LineStyle.SOLID
            The style of the edge.

        Returns
        -------
        Plot
            The current plot with the edge properties set.
        """
        self.markers.with_edge(color=color, alpha=alpha, width=width, style=style)
        return self

    def with_text(
        self,
        strings: list[str],
        *,
        color: ColorType = "black",
        size: float = 12,
        rotation: float = 0.0,
        anchor: str | Alignment = Alignment.BOTTOM_LEFT,
        family: str | None = None,
    ) -> LabeledPlot:
        """
        Add texts at the positions of markers.

        Parameters
        ----------
        strings : list of str
            The text to add to the markers.
        color : ColorType, default "black"
            The color of the text.
        size : float, default 12
            The size of the text.
        rotation : float, default 0.0
            The rotation of the text.
        anchor : str or Alignment, default Alignment.BOTTOM_LEFT
            The anchor position of the text.
        family : str or None, optional
            The font family of the text.

        Returns
        -------
        LabeledPlot
            The current plot with the text added to the markers.
        """
        from whitecanvas.layers.group.labeled import LabeledPlot

        strings = _text_utils.norm_label_text(strings, self.line.data)
        texts = Texts(
            *self.line.data,
            strings,
            color=color,
            size=size,
            rotation=rotation,
            anchor=anchor,
            family=family,
            backend=self._backend_name,
        )
        old_name = self.name
        self.name = f"plot-of-{old_name}"
        xerr = Errorbars.empty_v(f"xerr-of-{old_name}", backend=self._backend_name)
        yerr = Errorbars.empty_h(f"yerr-of-{old_name}", backend=self._backend_name)
        return LabeledPlot(self, xerr, yerr, texts, name=old_name)

    def with_xerr(
        self,
        err: ArrayLike,
        err_high: ArrayLike | None = None,
        *,
        color: ColorType | _Void = _void,
        width: float | _Void = _void,
        style: str | _Void = _void,
        antialias: bool | _Void = _void,
        capsize: float = 0,
    ) -> LabeledPlot:
        """
        Add horizontal error bars from the markers.

        Parameters
        ----------
        err : array-like
            The error values.
        err_high : array-like or None, optional
            The high error values. If not given, use the same values as `err`.
        color : ColorType or None, optional
            The color of the error bars. If not given, use the color of the line.
        width : float or None, optional
            The width of the error bars. If not given, use the width of the markers.
        style : str or None, optional
            The style of the error bars. If not given, use the style of the markers.
        antialias : bool or None, optional
            Whether to use antialiasing. If not given, use the antialiasing of the line.
        capsize : float, default 0
            The size of the caps at the end of the error bars.

        Returns
        -------
        LabeledPlot
            The current plot with the horizontal error bars added.
        """
        from whitecanvas.layers._primitive import Errorbars
        from whitecanvas.layers.group.labeled import LabeledPlot

        if err_high is None:
            err_high = err
        if color is _void:
            color = self.line.color
        if width is _void:
            width = self.markers.edge.width
        if style is _void:
            style = self.markers.edge.style
        if antialias is _void:
            antialias = self.line.antialias
        old_name = self.name
        self.name = f"plot-of-{old_name}"
        xerr = Errorbars(
            self.data.y, self.data.x - err, self.data.x + err_high, color=color,
            width=width, style=style, antialias=antialias, capsize=capsize,
            orient=Orientation.HORIZONTAL, name=f"xerr-of-{old_name}",
            backend=self._backend_name,
        )  # fmt: skip
        yerr = Errorbars.empty_v(f"yerr-of-{old_name}", backend=self._backend_name)
        return LabeledPlot(self, xerr, yerr, name=old_name)

    def with_yerr(
        self,
        err: ArrayLike,
        err_high: ArrayLike | None = None,
        *,
        color: ColorType | _Void = _void,
        width: float | _Void = _void,
        style: str | _Void = _void,
        antialias: bool | _Void = _void,
        capsize: float = 0,
    ) -> LabeledPlot:
        """
        Add vertical error bars from the markers.

        Parameters
        ----------
        err : array-like
            The error values.
        err_high : array-like or None, optional
            The high error values. If not given, use the same values as `err`.
        color : ColorType or None, optional
            The color of the error bars. If not given, use the color of the line.
        width : float or None, optional
            The width of the error bars. If not given, use the width of the markers.
        style : str or None, optional
            The style of the error bars. If not given, use the style of the markers.
        antialias : bool or None, optional
            Whether to use antialiasing. If not given, use the antialiasing of the line.
        capsize : float, default 0
            The size of the caps at the end of the error bars.

        Returns
        -------
        LabeledPlot
            The current plot with the vertical error bars added.
        """
        from whitecanvas.layers._primitive import Errorbars
        from whitecanvas.layers.group.labeled import LabeledPlot

        if err_high is None:
            err_high = err
        if color is _void:
            color = self.line.color
        if width is _void:
            width = self.markers.edge.width
        if style is _void:
            style = self.markers.edge.style
        if antialias is _void:
            antialias = self.line.antialias
        old_name = self.name
        self.name = f"plot-of-{old_name}"
        yerr = Errorbars(
            self.data.x, self.data.y - err, self.data.y + err_high, color=color,
            width=width, style=style, antialias=antialias, capsize=capsize,
            name=f"yerr-of-{old_name}", backend=self._backend_name,
        )  # fmt: skip
        xerr = Errorbars.empty_h(f"xerr-of-{old_name}", backend=self._backend_name)
        return LabeledPlot(self, xerr, yerr, name=old_name)

    @property
    def data(self):
        """The internal data of the line and markers."""
        return self.line.data

    def set_data(self, xdata=None, ydata=None):
        self.line.set_data(xdata, ydata)
        self.markers.set_data(xdata, ydata)

    def _as_legend_item(self):
        line = self.line._as_legend_item()
        markers = self.markers._as_legend_item()
        return _legend.PlotLegendItem(line, markers)



from whitecanvas import new_canvas

def main():
    canvas = new_canvas("plotly")
    line0 = canvas.add_line([0, 1, 0, 1], color="blue", width=3)
    line1 = canvas.add_line([1, 0, 1, 0], color="blue", width=3)
    markers = (
        canvas.add_markers([2, 2, 2, 2], color="blue", size=20)
        .with_face_multi()
    )
    @line0.events.clicked.connect
    def on_line_clicked(i: int):
        line0.color = "red"

    @line1.events.clicked.connect
    def on_line_clicked(i: int):
        line1.color = "red"

    @markers.events.clicked.connect
    def on_markers_clicked(i: int):
        colors = markers.face.color
        colors[i] = [1, 0, 0, 1]
        markers.face.color = colors
    canvas.show(block=True)

if __name__ == "__main__":
    main()



from whitecanvas import new_canvas
import pandas as pd

def main():
    url = "https://raw.githubusercontent.com/mwaskom/seaborn-data/master/tips.csv"
    df = pd.read_csv(url)

    canvas = new_canvas("matplotlib:qt")

    layer = (
        canvas.cat_x(df, "smoker", "tip")
        .add_violinplot(color="sex")
        .as_edge_only()
        .with_outliers(symbol="D")
    )
    canvas.add_legend()
    canvas.show(block=True)

if __name__ == "__main__":
    main()



from whitecanvas.layers._base import (
    DataBoundLayer,
    Layer,
    LayerGroup,
    LayerWrapper,
    PrimitiveLayer,
)
from whitecanvas.layers._ndim import LayerStack
from whitecanvas.layers._primitive import (
    Band,
    Bars,
    Errorbars,
    Image,
    InfCurve,
    InfLine,
    Line,
    LineStep,
    Markers,
    MultiLine,
    Rects,
    Rug,
    Spans,
    Texts,
    Vectors,
)

__all__ = [
    "Layer",
    "PrimitiveLayer",
    "DataBoundLayer",
    "LayerGroup",
    "LayerWrapper",
    "LayerStack",
    "Line",
    "LineStep",
    "MultiLine",
    "Markers",
    "Bars",
    "Spans",
    "Errorbars",
    "Band",
    "Rug",
    "InfLine",
    "InfCurve",
    "Texts",
    "Rects",
    "Image",
    "Vectors",
]



from __future__ import annotations

import warnings
from timeit import default_timer
from typing import Callable

import matplotlib as mpl
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.artist import Artist
from matplotlib.backend_bases import MouseButton as mplMouseButton
from matplotlib.backend_bases import MouseEvent as mplMouseEvent
from matplotlib.collections import Collection
from matplotlib.lines import Line2D

from whitecanvas import protocols
from whitecanvas.backend.matplotlib._base import MplLayer, MplMouseEventsMixin
from whitecanvas.backend.matplotlib._labels import (
    MplAxis,
    MplLabel,
    MplTicks,
    Title,
)
from whitecanvas.backend.matplotlib._legend import make_sample_item
from whitecanvas.backend.matplotlib.bars import Bars
from whitecanvas.backend.matplotlib.image import Image as whitecanvasImage
from whitecanvas.backend.matplotlib.text import Texts as whitecanvasText
from whitecanvas.layers._legend import LegendItem, LegendItemCollection
from whitecanvas.types import (
    Location,
    Modifier,
    MouseButton,
    MouseEvent,
    MouseEventType,
    Rect,
)


@protocols.check_protocol(protocols.CanvasProtocol)
class Canvas:
    def __init__(self, ax: plt.Axes):
        self._axes = ax
        self._xaxis = MplAxis(self, "x")
        self._yaxis = MplAxis(self, "y")
        self._title = Title(self)
        self._xlabel = MplLabel(self, "x")
        self._ylabel = MplLabel(self, "y")
        self._xticks = MplTicks(self, "x")
        self._yticks = MplTicks(self, "y")
        ax.set_axisbelow(True)  # grid lines below other layers
        self._annot = ax.annotate(
            text="", xy=(0, 0), xytext=(20, -20), textcoords="data",
            bbox={"fc": np.array([1.0, 1.0, 1.0, 0.67])},
            fontproperties={"size": 14, "family": "Arial"},
            clip_on=False, zorder=10000,
        )  # fmt: skip
        self._annot.set_visible(False)
        self._hoverable_artists: list[MplMouseEventsMixin] = []
        self._last_hover = -1.0

        fig = self._axes.figure
        if fig is None:
            return
        fig.canvas.mpl_connect("motion_notify_event", self._on_hover)
        fig.canvas.mpl_connect("figure_leave_event", self._hide_tooltip)

    def _on_hover(self, event: mplMouseEvent):
        if default_timer() - self._last_hover < 0.1:
            # avoid calling the tooltip too often
            return
        if event.button is not None:
            return self._hide_tooltip()
        self._last_hover = default_timer()
        if event.inaxes is not self._axes:
            return self._hide_tooltip()
        for layer in reversed(self._hoverable_artists):
            text = layer._on_hover(event)
            if text:
                xy = event.xdata, event.ydata
                self._set_tooltip(xy, text)
                return
        self._hide_tooltip()

    def _set_tooltip(self, pos: tuple[float, float], text: str):
        # determine in which direction to show the tooltip
        x, y = pos
        x0, x1 = self._axes.get_xlim()
        y0, y1 = self._axes.get_ylim()
        xc = (x0 + x1) / 2
        yc = (y0 + y1) / 2
        dx = (x1 - x0) * 0.05
        dy = (y1 - y0) * 0.05
        xtext = x - dx if x > xc else x + dx
        ytext = y - dy if y > yc else y + dy
        ha = "right" if x > xc else "left"
        va = "top" if y > yc else "bottom"

        # update the tooltip
        self._annot.xy = pos
        self._annot.set_text(text)
        self._annot.set_position((xtext, ytext))
        self._annot.set_verticalalignment(va)
        self._annot.set_horizontalalignment(ha)
        self._annot.set_visible(True)
        if fig := self._axes.get_figure():
            fig.canvas.draw_idle()

    def _hide_tooltip(self, *_):
        if self._annot.get_visible():
            self._annot.set_visible(False)
            if fig := self._axes.get_figure():
                fig.canvas.draw_idle()

    def _plt_get_native(self):
        return self._axes

    def _plt_get_title(self):
        return self._title

    def _plt_get_xaxis(self):
        return self._xaxis

    def _plt_get_yaxis(self):
        return self._yaxis

    def _plt_get_xlabel(self):
        return self._xlabel

    def _plt_get_ylabel(self):
        return self._ylabel

    def _plt_get_xticks(self):
        return self._xticks

    def _plt_get_yticks(self):
        return self._yticks

    def _plt_reorder_layers(self, layers: list[MplLayer]):
        for i, layer in enumerate(layers):
            layer._plt_set_zorder(i)
        self._hoverable_artists.sort(key=lambda a: a.get_zorder())

    def _plt_get_aspect_ratio(self) -> float | None:
        out = self._axes.get_aspect()
        if out == "auto":
            return None
        return out

    def _plt_set_aspect_ratio(self, ratio: float | None):
        if ratio is None:
            self._axes.set_aspect("auto")
        else:
            self._axes.set_aspect(ratio)

    def _plt_add_layer(self, layer: Artist):
        if isinstance(layer, Line2D):
            self._axes.add_line(layer)
        elif isinstance(layer, Collection):
            self._axes.add_collection(layer, autolim=False)
        elif isinstance(layer, Bars):
            for child in layer.patches:
                self._axes.add_patch(child)
            self._axes.add_container(layer)
        elif isinstance(layer, whitecanvasText):
            layer.set_transform(self._axes.transData)
            for child in layer.get_children():
                self._axes._add_text(child)
                child.set_clip_path(self._axes.patch)
        elif isinstance(layer, whitecanvasImage):
            self._axes.add_artist(layer)
        else:
            raise NotImplementedError(f"{layer}")
        if hasattr(layer, "post_add"):
            layer.post_add(self)
        if hasattr(layer, "_on_hover"):
            self._hoverable_artists.append(layer)

    def _plt_remove_layer(self, layer: Artist):
        """Remove layer from the canvas"""
        layer.remove()
        if layer in self._hoverable_artists:
            self._hoverable_artists.remove(layer)

    def _plt_get_visible(self) -> bool:
        """Get visibility of canvas"""
        return self._axes.get_visible()

    def _plt_set_visible(self, visible: bool):
        """Set visibility of canvas"""
        self._axes.set_visible(visible)

    def _plt_twinx(self) -> Canvas:
        axnew = self._axes.twinx()
        return Canvas(axnew)

    def _plt_twiny(self) -> Canvas:
        axnew = self._axes.twiny()
        return Canvas(axnew)

    def _plt_inset(self, rect: Rect) -> Canvas:
        axnew = self._axes.inset_axes(
            (rect.left, rect.bottom, rect.width, rect.height), zorder=1000
        )
        return Canvas(axnew)

    def _plt_connect_mouse_click(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""

        def _cb(ev: mplMouseEvent):
            if ev.inaxes is not self._axes or ev.dblclick:
                return
            callback(self._translate_mouse_event(ev, MouseEventType.PRESS))

        self._axes.figure.canvas.mpl_connect("button_press_event", _cb)

    def _plt_connect_mouse_drag(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""

        def _cb(ev: mplMouseEvent):
            if ev.inaxes is not self._axes or ev.dblclick:
                return
            callback(self._translate_mouse_event(ev, MouseEventType.MOVE))

        self._axes.figure.canvas.mpl_connect("motion_notify_event", _cb)

    def _plt_connect_mouse_double_click(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""

        def _cb(ev: mplMouseEvent):
            if ev.inaxes is not self._axes or not ev.dblclick:
                return
            callback(self._translate_mouse_event(ev, MouseEventType.DOUBLE_CLICK))

        self._axes.figure.canvas.mpl_connect("button_press_event", _cb)

    def _plt_connect_mouse_release(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""

        def _cb(ev: mplMouseEvent):
            if ev.inaxes is not self._axes or ev.dblclick:
                return
            callback(self._translate_mouse_event(ev, MouseEventType.RELEASE))

        self._axes.figure.canvas.mpl_connect("button_release_event", _cb)

    def _plt_draw(self):
        if fig := self._axes.get_figure():
            fig.canvas.draw_idle()

    def _plt_get_mouse_enabled(self):
        return self._axes.get_navigate()

    def _plt_set_mouse_enabled(self, enable: bool):
        if fig := self._axes.get_figure():
            if toolbar := getattr(fig.canvas, "toolbar", None):
                toolbar._update_buttons_checked()
                return
        return self._axes.set_navigate(enable)

    def _translate_mouse_event(
        self, ev: mplMouseEvent, typ: MouseEventType
    ) -> MouseEvent:
        if ev.key is None:
            modifiers = ()
        else:
            modifiers = []
            for k in ev.key.split("+"):
                if _MOUSE_MOD_MAP.get(k, None):
                    modifiers.append(_MOUSE_MOD_MAP.get(k, None))
            modifiers = tuple(modifiers)
        return MouseEvent(
            pos=(ev.xdata, ev.ydata),
            button=_MOUSE_BUTTON_MAP.get(ev.button, MouseButton.NONE),
            modifiers=modifiers,
            type=typ,
        )

    def _plt_connect_xlim_changed(
        self, callback: Callable[[tuple[float, float]], None]
    ):
        """Connect callback to x-limits changed event"""
        self._axes.callbacks.connect("xlim_changed", lambda ax: callback(ax.get_xlim()))

    def _plt_connect_ylim_changed(
        self, callback: Callable[[tuple[float, float]], None]
    ):
        """Connect callback to y-limits changed event"""
        self._axes.callbacks.connect("ylim_changed", lambda ax: callback(ax.get_ylim()))

    def _plt_make_legend(
        self,
        items: list[tuple[str, LegendItem]],
        anchor: Location = Location.TOP_RIGHT,
    ):
        artists: list[Artist] = []
        names: list[str] = []
        for name, item in items:
            if isinstance(item, LegendItemCollection):
                for _name, _it in item.items:
                    sample = make_sample_item(_it)
                    if sample is not None:
                        artists.append(sample)
                        names.append(_name)
            else:
                sample = make_sample_item(item)
                if sample is not None:
                    artists.append(sample)
                    names.append(name)
        if artists:
            loc, bbox_to_anchor = _LEGEND_LOC_MAP[anchor]
            font_size = self._plt_get_xticks()._plt_get_size()
            self._axes.legend(
                artists, names, loc=loc, bbox_to_anchor=bbox_to_anchor,
                prop={"size": font_size},
            )  # fmt: skip
            if anchor.is_side:
                self._axes.figure.tight_layout()


_MOUSE_BUTTON_MAP = {
    mplMouseButton.LEFT: MouseButton.LEFT,
    mplMouseButton.MIDDLE: MouseButton.MIDDLE,
    mplMouseButton.RIGHT: MouseButton.RIGHT,
    mplMouseButton.BACK: MouseButton.BACK,
    mplMouseButton.FORWARD: MouseButton.FORWARD,
}
_MOUSE_MOD_MAP = {
    "control": Modifier.CTRL,
    "ctrl": Modifier.CTRL,
    "shift": Modifier.SHIFT,
    "alt": Modifier.ALT,
    "meta": Modifier.META,
}
_LEGEND_LOC_MAP = {
    Location.TOP_RIGHT: ("upper right", None),
    Location.TOP_CENTER: ("upper center", None),
    Location.TOP_LEFT: ("upper left", None),
    Location.BOTTOM_RIGHT: ("lower right", None),
    Location.BOTTOM_CENTER: ("lower center", None),
    Location.BOTTOM_LEFT: ("lower left", None),
    Location.CENTER_RIGHT: ("center right", None),
    Location.CENTER_LEFT: ("center left", None),
    Location.CENTER: ("center", None),
    Location.TOP_SIDE_LEFT: ("lower left", (0, 1.03)),
    Location.TOP_SIDE_CENTER: ("lower center", (0.5, 1.03)),
    Location.TOP_SIDE_RIGHT: ("lower right", (1, 1.03)),
    Location.BOTTOM_SIDE_LEFT: ("upper left", (0, -0.03)),
    Location.BOTTOM_SIDE_CENTER: ("upper center", (0.5, -0.03)),
    Location.BOTTOM_SIDE_RIGHT: ("upper right", (1, -0.03)),
    Location.LEFT_SIDE_TOP: ("upper right", (-0.03, 1)),
    Location.LEFT_SIDE_CENTER: ("center right", (-0.03, 0.5)),
    Location.LEFT_SIDE_BOTTOM: ("lower right", (-0.03, 0)),
    Location.RIGHT_SIDE_TOP: ("upper left", (1.03, 1)),
    Location.RIGHT_SIDE_CENTER: ("center left", (1.03, 0.5)),
    Location.RIGHT_SIDE_BOTTOM: ("lower left", (1.03, 0)),
}


@protocols.check_protocol(protocols.CanvasGridProtocol)
class CanvasGrid:
    def __init__(self, heights: list[float], widths: list[float], app: str = "default"):
        nr, nc = len(heights), len(widths)
        self._gridspec = plt.GridSpec(
            nr, nc, height_ratios=heights, width_ratios=widths
        )
        if app == "qt":
            app = "QtAgg"
        elif app == "wx":
            app = "WXAgg"
        elif app == "gtk":
            app = "GTK3Agg"
        elif app == "tk":
            app = "TkAgg"
        elif app == "notebook":
            app = "nbAgg"
        if app != "default":
            mpl.use(app)
        self._fig = plt.figure()

    def _plt_add_canvas(self, row: int, col: int, rowspan: int, colspan: int) -> Canvas:
        r1 = row + rowspan
        c1 = col + colspan
        axes = self._fig.add_subplot(self._gridspec[row:r1, col:c1])
        axes.set_facecolor(self._fig.get_facecolor())
        return Canvas(axes)

    def _plt_add_canvas_3d(self, row: int, col: int, rowspan: int, colspan: int):
        from whitecanvas.backend.matplotlib.components3d import Canvas3D

        r1 = row + rowspan
        c1 = col + colspan
        axes = self._fig.add_subplot(self._gridspec[row:r1, col:c1], projection="3d")
        axes.set_facecolor(self._fig.get_facecolor())
        return Canvas3D(axes)

    def _plt_get_visible(self) -> bool:
        return self._fig.get_visible()

    def _plt_show(self):
        # TODO: show the inline plot again
        self._fig.show(warn=False)

    def _plt_get_background_color(self):
        return self._fig.get_facecolor()

    def _plt_set_background_color(self, color):
        self._fig.set_facecolor(color)
        for ax in self._fig.axes:
            ax.set_facecolor(color)

    def _plt_screenshot(self):
        import io

        fig = self._fig
        with io.BytesIO() as buff:
            fig.savefig(buff, format="raw")
            buff.seek(0)
            data = np.frombuffer(buff.getvalue(), dtype=np.uint8)
        # NOTE: fig.canvas.get_width_height() fails if the device pixel ratio is such
        # as 175%.
        wmax, hmax = fig.bbox.max
        width = int(round(wmax / fig.canvas.device_pixel_ratio))
        height = int(round(hmax / fig.canvas.device_pixel_ratio))
        img = data.reshape((height, width, -1))
        return img

    def _plt_set_figsize(self, width: int, height: int):
        dpi = self._fig.get_dpi()
        self._fig.set_size_inches(width / dpi, height / dpi)
        with warnings.catch_warnings():
            # if the size is small, tight_layout may raise a warning
            warnings.simplefilter("ignore")
            self._fig.tight_layout()

    def _plt_set_spacings(self, wspace: float, hspace: float):
        dpi = self._fig.get_dpi()
        nh, nw = self._gridspec.get_geometry()
        w_avg = self._fig.get_figwidth() / nw * dpi
        h_avg = self._fig.get_figheight() / nh * dpi
        self._gridspec.update(hspace=hspace / h_avg, wspace=wspace / w_avg)



from __future__ import annotations

import weakref
from typing import TYPE_CHECKING, Generic, Iterable, TypeVar

import numpy as np
from numpy.typing import NDArray

from whitecanvas._exceptions import ReferenceDeletedError

if TYPE_CHECKING:
    from whitecanvas.canvas import CanvasGrid

_T = TypeVar("_T")


class Animation:
    """
    Class for capturing animations.

    You can either manually capture frames using the `capture` method, or use
    this instance as a iterator: `anim(iterable)` or `anim.iter_range(num)`.

    Examples
    --------
    >>> from whitecanvas import plot as plt
    >>> from whitecanvas.animation import Animation
    >>> import numpy as np

    >>> ### Create a canvas and a line
    >>> canvas = plt.figure()
    >>> x = np.linspace(0, 4 * np.pi, 200)
    >>> line = plt.line(x, np.sin(x))

    >>> ### Create an animation
    >>> anim = Animation(canvas)
    >>> for i in anim.iter_range(20):
    ...     line.set_data(x, np.sin(x + i * np.pi / 10))

    >>> ### Save the animation
    >>> anim.save("animation.gif")
    """

    def __init__(self, grid: CanvasGrid):
        self._grid_ref = weakref.ref(grid)
        self._frames: list[NDArray[np.int8]] = []

    def __call__(self, iterable: Iterable[_T]) -> AnimationIterator[_T]:
        return AnimationIterator(self, iterable)

    def iter_range(self, *args) -> AnimationIterator[int]:
        """Equivalent to `anim(range(*args))`."""
        return AnimationIterator(self, range(*args))

    def _grid(self) -> CanvasGrid:
        grid = self._grid_ref()
        if grid is None:
            raise ReferenceDeletedError("CanvasGrid has been deleted.")
        return grid

    def capture(self):
        """Capture current canvas state."""
        grid = self._grid()
        self._frames.append(grid.screenshot())

    def save(self, filename: str, dt: float = 100, loop: int = 0):
        """Save animation to a file."""
        import imageio

        imageio.mimwrite(filename, self._frames, duration=dt, loop=loop)

    def asarray(self) -> NDArray[np.int8]:
        """Convert frames to a (N, X, Y, 4) numpy array."""
        return np.stack(self._frames, axis=0)


class AnimationIterator(Generic[_T]):
    def __init__(self, anim: Animation, iterable: Iterable[_T]):
        self._anim = anim
        self._iterable = iterable

    def __iter__(self):
        grid = self._anim._grid()
        for item in self._iterable:
            self._anim._frames.append(grid.screenshot())
            yield item
        self._anim._frames.append(grid.screenshot())



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Protocol, runtime_checkable

from whitecanvas.protocols.layer_protocols import BaseProtocol

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray

    from whitecanvas.layers._legend import LegendItem
    from whitecanvas.types import LineStyle, Location, MouseEvent


@runtime_checkable
class HasVisibility(Protocol):
    def _plt_get_visible(self) -> bool:
        """Get visibility of canvas"""

    def _plt_set_visible(self, visible: bool):
        """Set visibility of canvas"""


@runtime_checkable
class HasLayers(Protocol):
    def _plt_add_layer(self, layer: BaseProtocol):
        """Add layer to the canvas (no need for reordering)"""

    def _plt_remove_layer(self, layer: BaseProtocol):
        """Remove layer from the canvas"""

    def _plt_reorder_layers(self, layers: list[BaseProtocol]):
        """Reorder layers in the canvas"""


@runtime_checkable
class CanvasProtocol(HasVisibility, HasLayers, Protocol):
    def _plt_get_native(self) -> Any:
        """Get the native backend object."""

    def _plt_get_title(self) -> TextLabelProtocol:
        """Get title handler"""

    def _plt_get_xaxis(self) -> AxisProtocol:
        """Get x axis handler"""

    def _plt_get_yaxis(self) -> AxisProtocol:
        """Get y axis handler"""

    def _plt_get_xlabel(self) -> TextLabelProtocol:
        """Get x label handler"""

    def _plt_get_ylabel(self) -> TextLabelProtocol:
        """Get y label handler"""

    def _plt_get_xticks(self) -> TicksProtocol:
        """Get x ticks handler"""

    def _plt_get_yticks(self) -> TicksProtocol:
        """Get y ticks handler"""

    def _plt_get_aspect_ratio(self) -> float | None:
        """Get aspect ratio of canvas"""

    def _plt_set_aspect_ratio(self, ratio: float | None):
        """Set aspect ratio of canvas"""

    def _plt_connect_mouse_click(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""

    def _plt_connect_mouse_drag(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to drag event"""

    def _plt_connect_mouse_double_click(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to double-clicked event"""

    def _plt_connect_mouse_release(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to release event"""

    def _plt_connect_xlim_changed(
        self, callback: Callable[[tuple[float, float]], None]
    ):
        """Connect callback to x-limits changed event"""

    def _plt_connect_ylim_changed(
        self, callback: Callable[[tuple[float, float]], None]
    ):
        """Connect callback to y-limits changed event"""

    def _plt_draw(self):
        """Draw (update) canvas"""

    def _plt_make_legend(
        self,
        items: list[tuple[str, LegendItem]],
        anchor: Location,
    ):
        """Make legend for the given legend items"""

    def _plt_get_mouse_enabled(self) -> bool:
        """Get mouse interaction enabled state"""

    def _plt_set_mouse_enabled(self, enabled: bool):
        """Set mouse interaction enabled state"""


@runtime_checkable
class TextLabelProtocol(HasVisibility, Protocol):
    def _plt_get_text(self) -> Any:
        """Get text of label"""

    def _plt_set_text(self, text):
        """Set text of label"""

    def _plt_get_size(self) -> float:
        """Get size of text"""

    def _plt_set_size(self, size: float):
        """Set size of text"""

    def _plt_get_color(self) -> str:
        """Get color of text"""

    def _plt_set_color(self, color):
        """Set color of text"""

    def _plt_get_fontfamily(self) -> str:
        """Get font family of text"""

    def _plt_set_fontfamily(self, font):
        """Set font family of text"""


@runtime_checkable
class TicksProtocol(HasVisibility, Protocol):
    def _plt_get_tick_labels(self) -> tuple[list[float], list[str]]:
        """Get current tick labels"""

    def _plt_override_labels(self, pos: list[float], labels: list[str]):
        """Override tick labels"""

    def _plt_reset_override(self):
        """Reset tick label override"""

    def _plt_get_size(self) -> float:
        """Get size of text"""

    def _plt_set_size(self, size: float):
        """Set size of text"""

    def _plt_get_color(self) -> str:
        """Get color of text"""

    def _plt_set_color(self, color):
        """Set color of text"""

    def _plt_get_fontfamily(self) -> str:
        """Get font family of text"""

    def _plt_set_fontfamily(self, font):
        """Set font family of text"""

    def _plt_get_text_rotation(self) -> float:
        """Get rotation of text"""

    def _plt_set_text_rotation(self, rotation: float):
        """Set rotation of text"""


@runtime_checkable
class AxisProtocol(Protocol):
    def _plt_get_limits(self) -> tuple[float, float]:
        """Get limits of axis"""

    def _plt_set_limits(self, limits: tuple[float, float]):
        """Set limits of axis"""

    def _plt_get_color(self) -> np.ndarray:
        """Get color of axis"""

    def _plt_set_color(self, color):
        """Set color of axis"""

    def _plt_flip(self) -> None:
        """Flip axis"""

    def _plt_set_grid_state(self, visible: bool, color, width: float, style: LineStyle):
        """Set the grid line."""


@runtime_checkable
class AxisGridProtocol(Protocol):
    def _plt_get_visible(self) -> bool:
        """Get visibility of axis grid"""

    def _plt_set_visible(self, visible: bool):
        """Set visibility of axis grid"""

    def _plt_get_color(self) -> np.ndarray:
        """Get color of axis grid"""

    def _plt_set_color(self, color):
        """Set color of axis grid"""

    def _plt_get_width(self) -> float:
        """Get width of axis grid"""

    def _plt_set_width(self, width: float):
        """Set width of axis grid"""


@runtime_checkable
class CanvasGridProtocol(Protocol):
    def _plt_add_canvas(
        self, row: int, col: int, rowspan: int, colspan: int
    ) -> CanvasProtocol:
        """Add canvas to the grid"""

    def _plt_get_background_color(self):
        """Get background color of canvas"""

    def _plt_set_background_color(self, color):
        """Set background color of canvas"""

    def _plt_screenshot(self) -> NDArray[np.uint8]:
        """Get screenshot of canvas"""

    def _plt_show(self):
        """Show canvas"""

    def _plt_set_figsize(self, width: int, height: int):
        """Set size of canvas in pixels."""

    def _plt_set_spacings(self, wspace: float, hspace: float):
        """Set spacing between subplots"""



from __future__ import annotations

import numpy as np
from plotly import graph_objects as go

from whitecanvas.backend._not_implemented import face_hatch
from whitecanvas.backend.plotly._base import (
    PlotlyHoverableLayer,
    from_plotly_linestyle,
    to_plotly_linestyle,
)
from whitecanvas.protocols import BandProtocol, check_protocol
from whitecanvas.types import LineStyle, Orientation
from whitecanvas.utils.normalize import arr_color, rgba_str_color

# NOTE: plotly does not support hover text on the fill area.
# see https://github.com/plotly/plotly.py/issues/2399


@check_protocol(BandProtocol)
class Band(PlotlyHoverableLayer[go.Scatter]):
    def __init__(
        self,
        t: np.ndarray,
        ydata0: np.ndarray,
        ydata1: np.ndarray,
        orient: Orientation,
    ):
        if orient.is_vertical:
            x = np.concatenate([t, t[::-1]])
            y = np.concatenate([ydata0, ydata1[::-1]])
        else:
            x = np.concatenate([ydata0, ydata1[::-1]])
            y = np.concatenate([t, t[::-1]])
        self._props = {
            "x": x,
            "y": y,
            "mode": "lines",
            "fill": "toself",
            "fillcolor": "blue",
            # "hoverinfo": "skip", ... this also disables clicked event.
            "type": "scatter",
            "line": {"color": "blue", "width": 1, "dash": "solid", "simplify": False},
            "customdata": [""] * t.size,
            "hovertemplate": "%{customdata}<extra></extra>",
            "showlegend": False,
            "visible": True,
        }
        PlotlyHoverableLayer.__init__(self)

    ##### XYYDataProtocol #####
    def _plt_get_vertical_data(self):
        x = self._props["x"]
        y = self._props["y"]
        nx = len(x) // 2
        ny = len(y) // 2
        return x[:nx], y[:ny], y[ny:][::-1]

    def _plt_get_horizontal_data(self):
        x = self._props["x"]
        y = self._props["y"]
        nx = len(x) // 2
        ny = len(y) // 2
        return y[:ny], x[:nx], x[nx:][::-1]

    def _plt_set_vertical_data(self, t, ydata0, ydata1):
        x = np.concatenate([t, t[::-1]])
        y = np.concatenate([ydata0, ydata1[::-1]])
        self._props["x"] = x
        self._props["y"] = y

    def _plt_set_horizontal_data(self, t, ydata0, ydata1):
        x = np.concatenate([ydata0, ydata1[::-1]])
        y = np.concatenate([t, t[::-1]])
        self._props["x"] = x
        self._props["y"] = y

    def _plt_get_face_color(self):
        return arr_color(self._props["fillcolor"])

    def _plt_set_face_color(self, color):
        self._props["fillcolor"] = rgba_str_color(color)

    _plt_get_face_hatch, _plt_set_face_hatch = face_hatch()

    def _plt_get_edge_color(self):
        return arr_color(self._props["line"]["color"])

    def _plt_set_edge_color(self, color):
        self._props["line"]["color"] = rgba_str_color(color)

    def _plt_get_edge_width(self):
        return self._props["line"]["width"]

    def _plt_set_edge_width(self, width: float):
        self._props["line"]["width"] = width

    def _plt_get_edge_style(self):
        return from_plotly_linestyle(self._props["line"]["dash"])

    def _plt_set_edge_style(self, style: LineStyle):
        self._props["line"]["dash"] = to_plotly_linestyle(style)

    def _plt_get_antialias(self) -> bool:
        return not self._props["line"]["simplify"]

    def _plt_set_antialias(self, antialias: bool):
        self._props["line"]["simplify"] = not antialias

    def _plt_set_hover_text(self, text: str):
        self._hover_texts = [text]
        fig = self._fig_ref()
        if fig is not None:
            self._update_hover_texts(fig)

    def _update_hover_texts(self, fig: go.Figure):
        if self._hover_texts is None:
            return

        fig.update_traces(
            customdata=self._hover_texts * self._props["x"].size,
            selector={"uid": self._props["uid"]},
        )



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Iterable,
    Sequence,
    TypeVar,
    overload,
)

import numpy as np
from cmap import Color
from numpy.typing import NDArray

from whitecanvas import theme
from whitecanvas.backend import Backend
from whitecanvas.layers import _legend
from whitecanvas.layers import group as _lg
from whitecanvas.layers.tabular import _jitter, _shared
from whitecanvas.layers.tabular import _plans as _p
from whitecanvas.layers.tabular._df_compat import DataFrameWrapper
from whitecanvas.types import (
    ColormapType,
    ColorType,
    Hatch,
    LineStyle,
    Orientation,
    XYYData,
)
from whitecanvas.utils.collections import OrderedSet
from whitecanvas.utils.normalize import as_any_1d_array, as_color_array, parse_texts
from whitecanvas.utils.type_check import is_real_number

if TYPE_CHECKING:
    from typing_extensions import Self


_DF = TypeVar("_DF")


class AreaCollection(_lg.LayerCollectionBase[_lg.Area]):
    """Collection of lines."""

    _ATTACH_TO_AXIS = True

    def __init__(
        self,
        data: list[XYYData],
        name: str | None = None,
        orient: Orientation = Orientation.VERTICAL,
        fill_alpha: float = 0.2,
        backend: str | Backend | None = None,
    ):
        areas = list[_lg.Area]()
        ori = Orientation.parse(orient)
        for d in data:
            area = _lg.Area(d.x, d.ydiff, d.y0, orient=ori, backend=backend)
            area.fill_alpha = fill_alpha
            areas.append(area)
        super().__init__(areas, name=name)
        self._orient = ori
        self._fill_alpha = fill_alpha

    @property
    def fill_alpha(self) -> float:
        return self._fill_alpha

    @property
    def width(self) -> NDArray[np.float32]:
        return np.array([area.line.width for area in self], dtype=np.float32)

    @width.setter
    def width(self, width: float | Sequence[float]):
        if is_real_number(width):
            _width = [width] * len(self)
        else:
            _width = np.asarray(width, dtype=np.float32)
        if len(_width) != len(self):
            raise ValueError(
                f"width must be a float or a sequence of length {len(self)}"
            )
        for area, w in zip(self, _width):
            area.line.width = w

    @property
    def color(self) -> NDArray[np.float32]:
        return np.array([area.color for area in self], dtype=np.float32)

    @color.setter
    def color(self, color: str | Sequence[str]):
        col = as_color_array(color, len(self))
        for area, c in zip(self, col):
            area.color = c

    @property
    def style(self) -> list[LineStyle]:
        return np.array([area.line.style for area in self], dtype=object)

    @style.setter
    def style(self, style: str | Sequence[str]):
        styles = as_any_1d_array(style, len(self), dtype=object)
        for area, s in zip(self, styles):
            area.line.style = s

    @property
    def hatch(self) -> list[str]:
        return np.array([area.fill.face.hatch for area in self], dtype=object)

    @hatch.setter
    def hatch(self, hatch: str | Sequence[str]):
        hatches = as_any_1d_array(hatch, len(self), dtype=object)
        for area, h in zip(self, hatches):
            area.fill.face.hatch = h

    @property
    def orient(self) -> Orientation:
        """Orientation of the filling areas."""
        return self._orient

    def with_hover_texts(self, text: str | Iterable[Any]) -> Self:
        if isinstance(text, str):
            texts = [text] * len(self)
        else:
            texts = [str(t) for t in text]
            if len(texts) != len(self):
                raise ValueError("Length of texts must match the number of lines.")
        for area, txt in zip(self, texts):
            area.line.with_hover_text(txt)
            area.fill.with_hover_text(txt)
        return self

    def with_hover_template(
        self,
        template: str,
        extra: Any | None = None,
    ) -> Self:
        """Define hover template to the layer."""
        if self._backend_name in ("plotly", "bokeh"):  # conversion for HTML
            template = template.replace("\n", "<br>")
        params = parse_texts(template, len(self), extra)
        # set default format keys
        if "i" not in params:
            params["i"] = np.arange(len(self))
        texts = [
            template.format(**{k: v[i] for k, v in params.items()})
            for i in range(len(self))
        ]
        return self.with_hover_texts(texts)


class DFArea(_shared.DataFrameLayerWrapper[AreaCollection, _DF], Generic[_DF]):
    def __init__(
        self,
        source: DataFrameWrapper[_DF],
        data: list[XYYData],
        categories: list[tuple[Any, ...]],
        color: str | tuple[str, ...] | None = None,
        width: float = 1.0,
        style: str | tuple[str, ...] | None = None,
        hatch: str | tuple[str, ...] | None = None,
        orient: Orientation = Orientation.VERTICAL,
        fill_alpha: float = 0.2,
        stackby: str | tuple[str, ...] | None = None,
        name: str | None = None,
        backend: str | Backend | None = None,
    ):
        splitby = _shared.join_columns(color, style, hatch, source=source)
        self._color_by = _p.ColorPlan.default()
        self._style_by = _p.StylePlan.default()
        self._hatch_by = _p.HatchPlan.default()
        self._categories = categories
        self._splitby = splitby
        base = AreaCollection(
            data, name=name, orient=orient, fill_alpha=fill_alpha, backend=backend
        )
        super().__init__(base, source)
        if color is not None:
            self.update_color(color)
        self.update_width(width)
        if style is not None:
            self.update_style(style)
        if hatch is not None:
            self.update_hatch(hatch)
        self.with_hover_template("\n".join(f"{k}: {{{k}!r}}" for k in self._splitby))
        self._stackby = stackby

    @property
    def orient(self) -> Orientation:
        """Orientation of the filling areas."""
        return self.base.orient

    @classmethod
    def from_table_stacked(
        cls,
        df: DataFrameWrapper[_DF],
        x: str | _jitter.JitterBase,
        y: str | _jitter.JitterBase,
        stackby: str | tuple[str, ...] | None,
        *,
        color: str | tuple[str, ...] | None = None,
        width: float = 1.0,
        style: str | tuple[str, ...] | None = None,
        hatch: str | tuple[str, ...] | None = None,
        name: str | None = None,
        orient: Orientation = Orientation.VERTICAL,
        fill_alpha: float = 0.2,
        backend: str | Backend | None = None,
    ) -> DFArea[_DF]:
        splitby = _shared.join_columns(stackby, color, style, hatch, source=df)
        if stackby is None:
            stackby = splitby
        elif isinstance(stackby, str):
            stackby = (stackby,)
        nstackcol = len(stackby)
        if isinstance(x, _jitter.JitterBase):
            xj = x
        else:
            xj = _jitter.IdentityJitter(x)
        if isinstance(y, _jitter.JitterBase):
            yj = y
        else:
            yj = _jitter.IdentityJitter(y)
        # pre-calculate all the possible xs
        all_x = list(OrderedSet(xj.map(df)))
        x_to_i = {x: i for i, x in enumerate(all_x)}

        def _hash_rule(x: float) -> int:
            return int(round(x * 1000))

        stack_cat = [sl for sl, _ in df.group_by(splitby[nstackcol:])]
        ycumsum = {(_hash_rule(_x), *_s): 0.0 for _x in all_x for _s in stack_cat}
        data = list[XYYData]()
        categories = list[tuple[Any, ...]]()
        for sl, sub in df.group_by(splitby):
            this_h = np.zeros_like(all_x)
            for _x, _h in zip(xj.map(sub), yj.map(sub)):
                this_h[x_to_i[_x]] = _h
            bottom = []
            _stack_cat = sl[nstackcol:]
            for _x, _h in zip(all_x, this_h):
                _x_hash = _hash_rule(_x)
                dy = ycumsum[(_x_hash, *_stack_cat)]
                bottom.append(dy)
                ycumsum[(_x_hash, *_stack_cat)] += _h
            categories.append(sl)
            data.append(XYYData(all_x, bottom, this_h + bottom))
        return DFArea(
            df, data, categories=categories, color=color, hatch=hatch, width=width,
            stackby=stackby, name=name, orient=orient, fill_alpha=fill_alpha,
            backend=backend,
        )  # fmt: skip

    @overload
    def update_color(self, value: ColorType) -> Self:
        ...

    @overload
    def update_color(
        self,
        by: str | Iterable[str],
        palette: ColormapType | None = None,
    ) -> Self:
        ...

    def update_color(self, by, /, palette=None):
        """Update the color rule of the layer."""
        cov = _shared.ColumnOrValue(by, self._source)
        if cov.is_column:
            if set(cov.columns) > set(self._splitby):
                raise ValueError(f"Cannot color by a column other than {self._splitby}")
            color_by = _p.ColorPlan.from_palette(cov.columns, palette=palette)
        else:
            color_by = _p.ColorPlan.from_const(Color(cov.value))
        self._base_layer.color = color_by.generate(self._categories, self._splitby)
        self._color_by = color_by
        return self

    def update_width(self, value: float) -> Self:
        """Update width of the lines."""
        self._base_layer.width = value
        return self

    def update_style(self, by: str | Iterable[str], styles=None) -> Self:
        """Update the styling rule of the layer."""
        cov = _shared.ColumnOrValue(by, self._source)
        if cov.is_column:
            if set(cov.columns) > set(self._splitby):
                raise ValueError(f"Cannot style by a column other than {self._splitby}")
            style_by = _p.StylePlan.new(cov.columns, values=styles)
        else:
            style_by = _p.StylePlan.from_const(LineStyle(cov.value))
        self._base_layer.style = style_by.generate(self._categories, self._splitby)
        self._style_by = style_by
        return self

    def update_hatch(self, by: str | Iterable[str], hatches=None) -> Self:
        """Update the hatch rule of the layer."""
        cov = _shared.ColumnOrValue(by, self._source)
        if cov.is_column:
            if set(cov.columns) > set(self._splitby):
                raise ValueError(f"Cannot hatch by a column other than {self._splitby}")
            hatch_by = _p.HatchPlan.new(cov.columns, values=hatches)
        else:
            hatch_by = _p.HatchPlan.from_const(cov.value)
        self._base_layer.hatch = hatch_by.generate(self._categories, self._splitby)
        self._hatch_by = hatch_by
        for area in self.base:
            area.fill.edge.width = 1.0
        return self

    def move(self, dx: float = 0.0, dy: float = 0.0, autoscale: bool = True) -> Self:
        """Add a constant shift to the layer."""
        for layer in self._base_layer:
            old_data = layer.data
            new_data = old_data[0] + dx, old_data[1] + dy
            layer.data = new_data
        if autoscale and (canvas := self._canvas_ref()):
            canvas._autoscale_for_layer(self, pad_rel=0.025)
        return self

    def with_hover_template(self, template: str) -> Self:
        """Define hover template to the layer."""
        extra = {}
        for i, key in enumerate(self._splitby):
            extra[key] = [row[i] for row in self._categories]
        self.base.with_hover_template(template, extra=extra)
        return self

    def _prep_legend_info(
        self,
    ) -> tuple[
        list[tuple[str, ColorType]],
        list[tuple[str, LineStyle]],
        list[tuple[str, Hatch]],
    ]:
        df = _shared.list_to_df(self._categories, self._splitby)
        color_entries = self._color_by.to_entries(df)
        style_entries = self._style_by.to_entries(df)
        hatch_entries = self._hatch_by.to_entries(df)
        return color_entries, style_entries, hatch_entries

    def _as_legend_item(self) -> _legend.LegendItemCollection | _legend.LineLegendItem:
        colors, styles, hatches = self._prep_legend_info()
        ncolor = len(colors)
        nstyle = len(styles)
        nhatch = len(hatches)
        widths = self._base_layer.width

        color_default = theme.get_theme().foreground_color
        hatch_default = Hatch.SOLID
        style_default = LineStyle.SOLID
        if ncolor == 1:
            _, color_default = colors[0]
        if nstyle == 1:
            _, style_default = styles[0]
        if nhatch == 1:
            _, hatch_default = hatches[0]

        face_color_default = Color([*color_default.rgba[:3], self.base.fill_alpha])
        if ncolor == 1 and nstyle == 1 and nhatch == 1:
            face = _legend.FaceInfo(face_color_default, hatch_default)
            edge = _legend.EdgeInfo(color_default, widths[0], style_default)
            return _legend.BarLegendItem(face, edge)
        items = []
        if ncolor > 1:
            items.append((", ".join(self._color_by.by), _legend.TitleItem()))
            for (label, color), w in zip(colors, widths):
                fc = Color([*color.rgba[:3], self.base.fill_alpha])
                face = _legend.FaceInfo(fc, hatch_default)
                edge = _legend.EdgeInfo(color, w, style_default)
                item = _legend.BarLegendItem(face, edge)
                items.append((label, item))
        if nstyle > 1:
            items.append((", ".join(self._style_by.by), _legend.TitleItem()))
            for (label, style), w in zip(styles, widths):
                face = _legend.FaceInfo(face_color_default, hatch_default)
                edge = _legend.EdgeInfo(color_default, w, style)
                item = _legend.BarLegendItem(
                    face,
                    edge,
                )
                items.append((label, item))
        if nhatch > 1:
            items.append((", ".join(self._hatch_by.by), _legend.TitleItem()))
            for (label, hatch), w in zip(hatches, widths):
                face = _legend.FaceInfo(face_color_default, hatch)
                edge = _legend.EdgeInfo(color_default, w, style_default)
                item = _legend.BarLegendItem(face, edge)
                items.append((label, item))
        return _legend.LegendItemCollection(items)



import matplotlib.pyplot as plt
from whitecanvas.canvas import Canvas
import pytest
import pyqtgraph as pg

ALL_BACKENDS = ["mock", "matplotlib", "pyqtgraph", "plotly", "bokeh", "vispy"]

def pytest_addoption(parser):
    parser.addoption("--backend", default="all")

@pytest.fixture(scope="function", params=ALL_BACKENDS)
def backend(request: pytest.FixtureRequest):
    backend = request.config.getoption("--backend")
    if backend != "all":
        if isinstance(backend, str):
            request.param = backend
        else:
            backend = list(backend)
            request.param.param = backend
    yield request.param
    # TODO: how to skip tests if failed in mock backend?
    if request.param == "matplotlib":
        plt.close("all")
    elif request.param == "pyqtgraph":
        canvas = Canvas._CURRENT_INSTANCE
        if canvas is not None and isinstance(canvas.native, pg.PlotItem):
            canvas.native.update()
            canvas.native.resize(100, 150)



from __future__ import annotations

from whitecanvas.backend import Backend
from whitecanvas.canvas import Canvas, CanvasGrid, SingleCanvas
from whitecanvas.core import new_canvas, new_grid


def current_grid() -> CanvasGrid:
    """Return the current canvas grid."""
    grid = CanvasGrid._CURRENT_INSTANCE
    if grid is None:
        grid = new_canvas()._grid
    return grid


def current_canvas() -> Canvas:
    """Return the current canvas."""
    canvas = Canvas._CURRENT_INSTANCE
    if canvas is None:
        canvas = new_canvas()
    return canvas


def show(block: bool = False, flush: bool = True):
    """Show the current canvas."""
    current_grid().show(block=block)
    if flush:
        Canvas._CURRENT_INSTANCE = None


def subplots(
    nrows: int = 1,
    ncols: int = 1,
) -> CanvasGrid:
    """Create a new grid of subplots."""
    return new_grid(nrows, ncols).fill()


def figure(
    *,
    size: tuple[float, float] | None = None,
    palette: str | None = None,
) -> SingleCanvas:
    """Create a new grid of subplots."""
    return new_canvas(size=size, palette=palette)


def use(backend: str):
    """Set the backend to use."""
    return Backend(backend)  # update default
